#ifndef MATRIX_H
#define MATRIX_H

#include "H.h"

template <class typ> struct Matrix {
  unsigned m, n;     // dimensions
  Matrix(unsigned n1, unsigned n2) : m(n1), n(n2) { } 
  virtual void amux(typ d, typ *x, typ *y) const = 0;   // y +=d*Ax
  virtual void precond_apply(typ*) const = 0;     // apply preconditioner
  virtual ~Matrix() { }
};

template <class typ> struct HSymMatrix : public Matrix<typ>
{
  mblock<typ>** blcks;
  blcluster* blclTree;

  HSymMatrix(unsigned n, blcluster* tree = NULL) : Matrix<typ>(n, n),
    blclTree(tree) { blcks = NULL; }
  ~HSymMatrix() { freembls(blclTree, blcks); }

  void amux(typ d, typ* x, typ* y) const {
    multa_HSym_vec(d, blclTree, blcks, x, y);
  }

  void precond_apply(typ* x) const { }
};


template <class typ> struct HMatrix : public Matrix<typ>
{
  mblock<typ>** blcks;
  blcluster* blclTree;

  HMatrix(unsigned n1, unsigned n2, blcluster* tree = NULL) :
    Matrix<typ>(n1, n2),
    blclTree(tree) { blcks = NULL; }
  ~HMatrix() { freembls(blclTree, blcks); }

  void amux(typ d, typ* x, typ* y) const {
    multa_H_vec(d, blclTree, blcks, x, y);
  }

  void precond_apply(typ* x) const { }
};


#include "sparse.h"

struct CRSMatrix : public Matrix<double>
{
  double* A;
  unsigned *jA, *iA;

  CRSMatrix(unsigned n1, unsigned n2) : Matrix<double>(n1, n2) {
    A = NULL; jA = iA = NULL;
  }
  ~CRSMatrix() { delete [] A; delete [] jA; delete [] iA; }

  void amux(double d, double* x, double *y) const {
    amuxCRS(n, d, x, y, A, jA, iA);
  }

  void precond_apply(double* x) const { }
};


struct CCSMatrix : public Matrix<double>
{
  double* A;
  unsigned *iA, *jA;

  CCSMatrix(unsigned n1, unsigned n2) : Matrix<double>(n1, n2) {
    A = NULL; iA = jA = NULL;
  }
  ~CCSMatrix() { delete [] A; delete [] iA; delete [] jA; }

  void amux(double d, double* x, double *y) const {
    atmuxCRS(n, d, x, y, A, iA, jA);  // not tested
  }

  void precond_apply(double* x) const { }
};

#endif
