#ifndef BLAS_H
#define BLAS_H

#include <assert.h>
#include <stdlib.h>
#include "complex.h"    //Complex arthimetic
//using namespace std;
#define SIGN(a) ((a) >= 0 ? 1.0 : -1.0)
#define SQR(a) ((a)*(a))                             // Quadrat von a
inline double abs2(double a) { return SQR(a); }
#define MIN(a,b) ((a) <= (b) ? (a) : (b))
#define MAX(a,b) ((a) >= (b) ? (a) : (b))
#define LTR(i,j,n) (((2*n-j-1)*j)/2+i)
#define UTR(i,j) ((j*(j+1))/2+i)
#define GE(i,j,n) ((i+j*n))

const double D_ZERO =  0.0;
const double D_ONE  =  1.0;
const double D_MONE = -1.0;
const float S_ZERO  =  0.0;
const float S_ONE   =  1.0;
const float S_MONE  = -1.0;
const scomp C_ZERO  = 0.0;
const scomp C_ONE   = 1.0;
const scomp C_MONE  = -1.0;
const dcomp Z_ZERO  = 0.0;
const dcomp Z_ONE   = 1.0;
const dcomp Z_MONE  = -1.0;

const double D_PREC = 1e-16;

const unsigned N_ONE = 1;
const char JOB_STR[] = "NTOSVULC";


#ifdef WIN32
  #define dcopy_ dcopy
  #define dnrm2_ dnrm2
  #define dgemv_ dgemv
  #define dspmv_ dspmv
  #define dtpmv_ dtpmv
  #define dgemm_ dgemm
  #define daxpy_ daxpy
  #define ddot_ ddot
  #define dscal_ dscal
  #define dgesvd_ dgesvd
  #define dgeqrf_ dgeqrf
  #define dormqr_ dormqr
  #define dorgqr_ dorgqr
  #define dgetri_ dgetri
  #define dgetrf_ dgetrf
  #define dsptri_ dsptri
  #define dsptrf_ dsptrf
  #define dtptrs_ dtptrs
  #define dpptrf_ dpptrf
  #define scopy_ scopy
  #define snrm2_ snrm2
  #define sgemv_ sgemv
  #define sspmv_ sspmv
  #define stpmv_ stpmv
  #define sgemm_ sgemm
  #define saxpy_ saxpy
  #define sdot_ sdot
  #define sscal_ sscal
  #define sgesvd_ sgesvd
  #define sgeqrf_ sgeqrf
  #define sormqr_ sormqr
  #define sorgqr_ sorgqr
  #define sgetri_ sgetri
  #define sgetrf_ sgetrf
  #define ssptri_ ssptri
  #define ssptrf_ ssptrf
  #define stptrs_ stptrs
  #define spptrf_ spptrf
  #define ccopy_ ccopy
  #define scnrm2_ scnrm2
  #define cgemv_ cgemv
  #define chpmv_ chpmv
  #define ctpmv_ ctpmv
  #define cgemm_ cgemm
  #define caxpy_ caxpy
  #define cdotc_ cdotc
  #define cscal_ cscal
  #define cgesvd_ cgesvd
  #define cgeqrf_ cgeqrf
  #define cunmqr_ cunmqr
  #define cungqr_ cungqr
  #define cgetri_ cgetri
  #define cgetrf_ cgetrf
  #define csptri_ csptri
  #define csptrf_ csptrf
  #define ctptrs_ ctptrs
  #define cpptrf_ cpptrf
  #define zcopy_ zcopy
  #define dznrm2_ dznrm2
  #define zgemv_ zgemv
  #define zhpmv_ zhpmv
  #define ztpmv_ ztpmv
  #define zgemm_ zgemm
  #define zaxpy_ zaxpy
  #define zdotc_ zdotc
  #define zscal_ zscal
  #define zgesvd_ zgesvd
  #define zgeqrf_ zgeqrf
  #define zunmqr_ zunmqr
  #define zungqr_ zungqr
  #define zgetri_ zgetri
  #define zgetrf_ zgetrf
  #define zsptri_ zsptri
  #define zsptrf_ zsptrf
  #define ztptrs_ ztptrs
  #define zpptrf_ zpptrf
#endif



extern "C" {
/******************************************************************/
//double precision real
/******************************************************************/
  unsigned idamax_(const unsigned*, const double*, const unsigned*);
  void dcopy_(const unsigned*, const double*, const unsigned*,
			double*, const unsigned*);
  void daxpy_(const unsigned*, const double*, double*,
			const unsigned*, double*, const unsigned*);
  void dscal_(const unsigned*, const double*, const double*,
			const unsigned*);
  double ddot_(const unsigned*, const double*, const unsigned*,
			 const double*, const unsigned*);
  double dnrm2_(const unsigned*, const double*, const unsigned*);

  void dgemm_(const char*, const char*, const unsigned*, const unsigned*,
	      const unsigned*, const double*, double*, const unsigned*,
	      double*, const unsigned*, const double*, double*,
	      const unsigned*);
  void dger_(const unsigned*, const unsigned*, const double*, const double*,
	     const unsigned*, double*, const unsigned*, const double*,
	     const unsigned*);
  void dgemv_(const char*, const unsigned*, const unsigned*, const double*,
	      const double*, const unsigned*, const double*, const unsigned*,
	      const double*, double*, const unsigned*);
  void dorgqr_(const unsigned*, const unsigned*, const unsigned*,
	       double*, const unsigned*, double*, double*,
	       const unsigned*, int*);
  void dormqr_(const char*, const char*, const unsigned*, const unsigned*,
	       const unsigned*, double*, const unsigned*, double*,
	       double*, const unsigned*, double*, const unsigned*,
	       int*);
  void dsyev_(const char*, const char*, const unsigned*, double*,
	     const unsigned*, double*, double*, const unsigned*, int*);
  void dgeqrf_(const unsigned*, const unsigned*, double*, const unsigned*,
	       double*, double*, const unsigned*, int*);
  void dgesvd_(const char*, const char*, const unsigned*, const unsigned*,
	       double*, const unsigned*, double*, double*, const unsigned*,
	       double*, const unsigned*, double*, const unsigned*, int*);
  void dgetrf_(const unsigned*, const unsigned*, double*, const unsigned*,
	       unsigned*, int*);
  void dgetri_(const unsigned*, double*, const unsigned*, unsigned*, double*,
	       const unsigned*, int*);
  void dspmv_(const char*, const unsigned*, const double*,
	      const double*, const double*, const unsigned*,
	      const double*, double*, const unsigned*);
  void dsptrf_(const char*, const unsigned*, const double*, const unsigned*,
	       int*);
  void dsptri_(const char*, const unsigned*, const double*, const unsigned*,
	       double*, int*);
  void dpotrf_(const char*, const unsigned*, const double*, const unsigned*,
	       int*);
  void dpotri_(const char*, const unsigned*, const double*, const unsigned*,
	       int*);
  void dpptrf_(const char*, const unsigned*, const double*, int*);
  void dpptri_(const char*, const unsigned*, const double*, int*);
  void dtptrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, double*, double*, const unsigned*, int*);

  void dtrtrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, const double*, const unsigned*,
	       double*, const unsigned*, int*);
  void dtpmv_(const char*, const char*, const char*, const unsigned*,
	      const double*, double*, const unsigned*);

/******************************************************************/
//single precision real
/******************************************************************/
  unsigned isamax_(const unsigned*, const float*, const unsigned*);
  void scopy_(const unsigned*, const float*, const unsigned*,
			float*, const unsigned*);
  void saxpy_(const unsigned*, const float*, float*,
			const unsigned*, float*, const unsigned*);
  void sscal_(const unsigned*, const float*, const float*,
			const unsigned*);
  float sdot_(const unsigned*, const float*, const unsigned*,
			 const float*, const unsigned*);
  float snrm2_(const unsigned*, const float*, const unsigned*);

  void sgemm_(const char*, const char*, const unsigned*, const unsigned*,
	      const unsigned*, const float*, float*, const unsigned*,
	      float*, const unsigned*, const float*, float*,
	      const unsigned*);
  void sger_(const unsigned*, const unsigned*, const float*, const float*,
	     const unsigned*, float*, const unsigned*, const float*,
	     const unsigned*);
  void sgemv_(const char*, const unsigned*, const unsigned*, const float*,
	      const float*, const unsigned*, const float*, const unsigned*,
	      const float*, float*, const unsigned*);
  void sorgqr_(const unsigned*, const unsigned*, const unsigned*,
	       float*, const unsigned*, float*, float*,
	       const unsigned*, int*);
  void sormqr_(const char*, const char*, const unsigned*, const unsigned*,
	       const unsigned*, float*, const unsigned*, float*,
	       float*, const unsigned*, float*, const unsigned*,
	       int*);
  void ssyev_(const char*, const char*, const unsigned*, float*,
	     const unsigned*, float*, float*, const unsigned*, int*);
  void sgeqrf_(const unsigned*, const unsigned*, float*, const unsigned*,
	       float*, float*, const unsigned*, int*);
  void sgesvd_(const char*, const char*, const unsigned*, const unsigned*,
	       float*, const unsigned*, float*, float*, const unsigned*,
	       float*, const unsigned*, float*, const unsigned*, int*);
  void sgetrf_(const unsigned*, const unsigned*, float*, const unsigned*,
	       unsigned*, int*);
  void sgetri_(const unsigned*, float*, const unsigned*, unsigned*, float*,
	       const unsigned*, int*);
  void sspmv_(const char*, const unsigned*, const float*,
	      const float*, const float*, const unsigned*,
	      const float*, float*, const unsigned*);
  void ssptrf_(const char*, const unsigned*, const float*, const unsigned*,
	       int*);
  void ssptri_(const char*, const unsigned*, const float*, const unsigned*,
	       float*, int*);
  void spotrf_(const char*, const unsigned*, const float*, const unsigned*,
	       int*);
  void spotri_(const char*, const unsigned*, const float*, const unsigned*,
	       int*);
  void spptrf_(const char*, const unsigned*, const float*, int*);
  void spptri_(const char*, const unsigned*, const float*, int*);
  void stptrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, float*, float*, const unsigned*, int*);

  void strtrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, const float*, const unsigned*,
	       float*, const unsigned*, int*);
  void stpmv_(const char*, const char*, const char*, const unsigned*,
	      const float*, float*, const unsigned*);
	      
/******************************************************************/
//single precision complex
/******************************************************************/

  void ccopy_(const unsigned*, const scomp*, const unsigned*,
			scomp*, const unsigned*);
  void caxpy_(const unsigned*, const scomp*, scomp*,
			const unsigned*, scomp*, const unsigned*);
  void cscal_(const unsigned*, const scomp*, const scomp*,
			const unsigned*);
  void cdotc_(const scomp*, const unsigned*, const scomp*, const unsigned*,
			 const scomp*, const unsigned*);
  float scnrm2_(const unsigned*, const scomp*, const unsigned*);

  void cgemm_(const char*, const char*, const unsigned*, const unsigned*,
	      const unsigned*, const scomp*, scomp*, const unsigned*,
	      scomp*, const unsigned*, const scomp*, scomp*,
	      const unsigned*);
  void cger_(const unsigned*, const unsigned*, const scomp*, const scomp*,
	     const unsigned*, scomp*, const unsigned*, const scomp*,
	     const unsigned*);
  void cgemv_(const char*, const unsigned*, const unsigned*, const scomp*,
	      const scomp*, const unsigned*, const scomp*, const unsigned*,
	      const scomp*, scomp*, const unsigned*);
  void cungqr_(const unsigned*, const unsigned*, const unsigned*,
	       scomp*, const unsigned*, scomp*, scomp*,
	       const unsigned*, int*);
  void cunmqr_(const char*, const char*, const unsigned*, const unsigned*,
	       const unsigned*, scomp*, const unsigned*, scomp*,
	       scomp*, const unsigned*, scomp*, const unsigned*,
	       int*);
  void csyev_(const char*, const char*, const unsigned*, scomp*,
	     const unsigned*, scomp*, scomp*, const unsigned*, int*);
  void cgeqrf_(const unsigned*, const unsigned*, scomp*, const unsigned*,
	       scomp*, scomp*, const unsigned*, int*);
  void cgesvd_(const char*, const char*, const unsigned*, const unsigned*,
	       scomp*, const unsigned*, float*, scomp*, const unsigned*,
	       scomp*, const unsigned*, scomp*, const unsigned*, float*, int*);
  void cgetrf_(const unsigned*, const unsigned*, scomp*, const unsigned*,
	       unsigned*, int*);
  void cgetri_(const unsigned*, scomp*, const unsigned*, unsigned*, scomp*,
	       const unsigned*, int*);
  void chpmv_(const char*, const unsigned*, const scomp*,
	      const scomp*, const scomp*, const unsigned*,
	      const scomp*, scomp*, const unsigned*);
  void csptrf_(const char*, const unsigned*, const scomp*, const unsigned*,
	       int*);
  void csptri_(const char*, const unsigned*, const scomp*, const unsigned*,
	       scomp*, int*);
  void cpotrf_(const char*, const unsigned*, const scomp*, const unsigned*,
	       int*);
  void cpotri_(const char*, const unsigned*, const scomp*, const unsigned*,
	       int*);
  void cpptrf_(const char*, const unsigned*, const scomp*, int*);
  void cpptri_(const char*, const unsigned*, const scomp*, int*);
  void ctptrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, scomp*, scomp*, const unsigned*, int*);

  void ctrtrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, const scomp*, const unsigned*,
	       scomp*, const unsigned*, int*);
  void ctpmv_(const char*, const char*, const char*, const unsigned*,
	      const scomp*, scomp*, const unsigned*);
	      
/******************************************************************/
//double precision complex
/******************************************************************/
  void zcopy_(const unsigned*, const dcomp*, const unsigned*,
			dcomp*, const unsigned*);
  void zaxpy_(const unsigned*, const dcomp*, dcomp*,
			const unsigned*, dcomp*, const unsigned*);
  void zscal_(const unsigned*, const dcomp*, const dcomp*,
			const unsigned*);
  void zdotc_(const dcomp*, const unsigned*, const dcomp*, const unsigned*,
			 const dcomp*, const unsigned*);
  double dznrm2_(const unsigned*, const dcomp*, const unsigned*);

  void zgemm_(const char*, const char*, const unsigned*, const unsigned*,
	      const unsigned*, const dcomp*, dcomp*, const unsigned*,
	      dcomp*, const unsigned*, const dcomp*, dcomp*,
	      const unsigned*);
  void zger_(const unsigned*, const unsigned*, const dcomp*, const dcomp*,
	     const unsigned*, dcomp*, const unsigned*, const dcomp*,
	     const unsigned*);
  void zgemv_(const char*, const unsigned*, const unsigned*, const dcomp*,
	      const dcomp*, const unsigned*, const dcomp*, const unsigned*,
	      const dcomp*, dcomp*, const unsigned*);
  void zungqr_(const unsigned*, const unsigned*, const unsigned*,
	       dcomp*, const unsigned*, dcomp*, dcomp*,
	       const unsigned*, int*);
  void zunmqr_(const char*, const char*, const unsigned*, const unsigned*,
	       const unsigned*, dcomp*, const unsigned*, dcomp*,
	       dcomp*, const unsigned*, dcomp*, const unsigned*,
	       int*);
  void zsyev_(const char*, const char*, const unsigned*, dcomp*,
	     const unsigned*, dcomp*, dcomp*, const unsigned*, int*);
  void zgeqrf_(const unsigned*, const unsigned*, dcomp*, const unsigned*,
	       dcomp*, dcomp*, const unsigned*, int*);
  void zgesvd_(const char*, const char*, const unsigned*, const unsigned*,
	       dcomp*, const unsigned*, double*, dcomp*, const unsigned*,
	       dcomp*, const unsigned*, dcomp*, const unsigned*, double*,
	       int*);
  void zgetrf_(const unsigned*, const unsigned*, dcomp*, const unsigned*,
	       unsigned*, int*);
  void zgetri_(const unsigned*, dcomp*, const unsigned*, unsigned*, dcomp*,
	       const unsigned*, int*);
  void zhpmv_(const char*, const unsigned*, const dcomp*,
	      const dcomp*, const dcomp*, const unsigned*,
	      const dcomp*, dcomp*, const unsigned*);
  void zsptrf_(const char*, const unsigned*, const dcomp*, const unsigned*,
	       int*);
  void zsptri_(const char*, const unsigned*, const dcomp*, const unsigned*,
	       dcomp*, int*);
  void zpotrf_(const char*, const unsigned*, const dcomp*, const unsigned*,
	       int*);
  void zpotri_(const char*, const unsigned*, const dcomp*, const unsigned*,
	       int*);
  void zpptrf_(const char*, const unsigned*, const dcomp*, int*);
  void zpptri_(const char*, const unsigned*, const dcomp*, int*);
  void ztptrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, dcomp*, dcomp*, const unsigned*, int*);

  void ztrtrs_(const char*, const char*, const char*, const unsigned*,
	       const unsigned*, const dcomp*, const unsigned*,
	       dcomp*, const unsigned*, int*);
  void ztpmv_(const char*, const char*, const char*, const unsigned*,
	      const dcomp*, dcomp*, const unsigned*);
}



namespace blas {

  inline unsigned maxi(const unsigned n, double* const v)
  { return idamax_(&n, v, &N_ONE); }
  inline unsigned maxi(const unsigned n, float* const v)
  { return isamax_(&n, v, &N_ONE); }

  inline void load(const unsigned n, double e, double* const v)
  { for (unsigned i=0; i<n; ++i) v[i] = e; }
  inline void load(const unsigned n, float e, float* const v)
  { for (unsigned i=0; i<n; ++i) v[i] = e; }
  inline void load(const unsigned n, scomp e, scomp* const v)
  { for (unsigned i=0; i<n; ++i) v[i]= e;}
  inline void load(const unsigned n, dcomp e, dcomp* const v)
  { for (unsigned i=0; i<n; ++i) v[i]= e;}
  inline void setzero(const unsigned n, double* const v)
  { load(n, D_ZERO, v); }
  inline void setzero(const unsigned n, float* const v)
  { load(n, S_ZERO, v); }
  inline void setzero(const unsigned n, scomp* const v)
  { load(n, C_ZERO, v); }
  inline void setzero(const unsigned n, dcomp* const v)
  { load(n, Z_ZERO, v); }
  
  inline double nrm2(const unsigned n, double* const v)
  { return dnrm2_(&n, v, &N_ONE); }
  inline float nrm2(const unsigned n, float* const v)
  { return snrm2_(&n, v, &N_ONE); }
  inline float nrm2(const unsigned n, scomp* const v)
  { return scnrm2_(&n, v, &N_ONE); }
  inline double nrm2(const unsigned n, dcomp* const v)
  { return dznrm2_(&n, v, &N_ONE); }

  inline void copy(const unsigned n, double* orig, double* dest)
  { dcopy_(&n, orig, &N_ONE, dest, &N_ONE); }
  inline void copy(const unsigned n, float* orig, float* dest)
  { scopy_(&n, orig, &N_ONE, dest, &N_ONE); }
  inline void copy(const unsigned n, scomp* orig, scomp* dest)
  { ccopy_(&n, orig, &N_ONE, dest, &N_ONE); }
  inline void copy(const unsigned n, dcomp* orig, dcomp* dest)
  { zcopy_(&n, orig, &N_ONE, dest, &N_ONE); }


  inline void copy(const unsigned n, double* orig, const unsigned inco,
		   double* dest, const unsigned incd)
  { dcopy_(&n, orig, &inco, dest, &incd); }
  inline void copy(const unsigned n, float* orig, const unsigned inco,
		   float* dest, const unsigned incd)
  { scopy_(&n, orig, &inco, dest, &incd); }
  inline void copy(const unsigned n, scomp* orig, const unsigned inco,
		   scomp* dest, const unsigned incd)
  { ccopy_(&n, orig, &inco, dest, &incd); }
  inline void copy(const unsigned n, dcomp* orig, const unsigned inco,
		   dcomp* dest, const unsigned incd)
  { zcopy_(&n, orig, &inco, dest, &incd); }
  
  // Conv.Copy double2float
  inline void copy(const unsigned n, double* orig, float* dest)
  { for (unsigned i=0; i<n; i++) dest[i] = (float) orig[i]; }

  // Conv.Copy float2double
  inline void copy(const unsigned n, float* orig, double* dest)
  { for (unsigned i=0; i<n; i++) dest[i] = (double) orig[i]; }

  // Conv.Copy dcomp2scomp
  inline void copy(const unsigned n, dcomp* orig, scomp* dest)
  { for (unsigned i=0; i<n; i++) 
      dest[i].re = (float) orig[i].re, dest[i].im = (float) orig[i].im; }

  // Conv.Copy scomp2dcomp
  inline void copy(const unsigned n, scomp* orig, dcomp* dest)
  { for (unsigned i=0; i<n; i++) 
      dest[i].re = (double) orig[i].re, dest[i].im = (double) orig[i].im; }

  inline double scpr(const unsigned n, double* const v1, double* const v2)
  { return ddot_(&n, v1, &N_ONE, v2, &N_ONE); }
  inline float scpr(const unsigned n, float* const v1, float* const v2)
  { return sdot_(&n, v1, &N_ONE, v2, &N_ONE); }
  inline scomp scpr(const unsigned n, scomp* const v1, scomp* const v2)
  { scomp z; cdotc_(&z, &n, v1, &N_ONE, v2, &N_ONE); return z; }
  inline dcomp scpr(const unsigned n, dcomp* const v1, dcomp* const v2)
  { dcomp z; zdotc_(&z, &n, v1, &N_ONE, v2, &N_ONE); return z;}


  inline void add(const unsigned n, double* const x, double* const y)
  { daxpy_(&n, &D_ONE, x, &N_ONE, y, &N_ONE); }
  inline void add(const unsigned n, float* const x, float* const y)
  { saxpy_(&n, &S_ONE, x, &N_ONE, y, &N_ONE); }
  inline void add(const unsigned n, scomp* const x, scomp* const y)
  { caxpy_(&n, &C_ONE, x, &N_ONE, y, &N_ONE); }
  inline void add(const unsigned n, dcomp* const x,dcomp* const y)
  { zaxpy_(&n, &Z_ONE, x, &N_ONE, y, &N_ONE); }

  inline void axpy(const unsigned n, const double d, double* const x,
		   double* const y) { daxpy_(&n, &d, x, &N_ONE, y, &N_ONE); }
  inline void axpy(const unsigned n, const float d, float* const x,
		   float* const y) { saxpy_(&n, &d, x, &N_ONE, y, &N_ONE); }
  inline void axpy(const unsigned n, const scomp d, scomp* const x,
		   scomp* const y) { caxpy_(&n, &d, x, &N_ONE, y, &N_ONE); }
  inline void axpy(const unsigned n, const dcomp d, dcomp* const x,
		   dcomp* const y) { zaxpy_(&n, &d, x, &N_ONE, y, &N_ONE); }

  inline void scal(const unsigned n, const double d, double* const x)
  { dscal_(&n, &d, x, &N_ONE); }
  inline void scal(const unsigned n, const float d, float* const x)
  { sscal_(&n, &d, x, &N_ONE); }
  inline void scal(const unsigned n, const scomp d, scomp* const x)
  { cscal_(&n, &d, x, &N_ONE); }
  inline void scal(const unsigned n, const dcomp d, dcomp* const x)
  { zscal_(&n, &d, x, &N_ONE); }

  template<class T>
  inline void scal(const unsigned n, const T d,
		   T* const x, T* const y)
  { for  (unsigned i=0; i<n; ++i) y[i] = d * x[i]; }

  // y += d Ax
  inline void gemva(const unsigned m, const unsigned n, double d, double* A,
		    double *x, double *y)
  { dgemv_(JOB_STR, &m, &n, &d, A, &m, x, &N_ONE, &D_ONE, y, &N_ONE); }
  inline void gemva(const unsigned m, const unsigned n, float d, float* A,
		    float *x, float *y)
  { sgemv_(JOB_STR, &m, &n, &d, A, &m, x, &N_ONE, &S_ONE, y, &N_ONE); }
  inline void gemva(const unsigned m, const unsigned n, scomp d, scomp* A,
		    scomp *x, scomp *y)
  { cgemv_(JOB_STR, &m, &n, &d, A, &m, x, &N_ONE, &C_ONE, y, &N_ONE); }
  inline void gemva(const unsigned m, const unsigned n, dcomp d, dcomp* A,
		    dcomp *x, dcomp *y)
  { zgemv_(JOB_STR, &m, &n, &d, A, &m, x, &N_ONE, &Z_ONE, y, &N_ONE); }

  // y += d A^T x
  inline void gemtva(const unsigned m, const unsigned n, double d, double* A,
		    double *x, double *y)
  { dgemv_(JOB_STR+1, &m, &n, &d, A, &m, x, &N_ONE, &D_ONE, y, &N_ONE); }
  inline void gemtva(const unsigned m, const unsigned n, float d, float* A,
		    float *x, float *y)
  { sgemv_(JOB_STR+1, &m, &n, &d, A, &m, x, &N_ONE, &S_ONE, y, &N_ONE); }
  inline void gemtva(const unsigned m, const unsigned n, scomp d, scomp* A,
		    scomp *x, scomp *y)
  { cgemv_(JOB_STR+7, &m, &n, &d, A, &m, x, &N_ONE, &C_ONE, y, &N_ONE); }
  inline void gemtva(const unsigned m, const unsigned n, dcomp d, dcomp* A,
		    dcomp *x, dcomp *y)
  { zgemv_(JOB_STR+7, &m, &n, &d, A, &m, x, &N_ONE, &Z_ONE, y, &N_ONE); }  
  
  // y += d A x (A symm. dense in packed format)
  inline void gemsva(const unsigned n, double d, double* A,
		    double *x, double *y)
  { dspmv_(JOB_STR+5, &n, &d, A, x, &N_ONE, &D_ONE, y, &N_ONE); }
  inline void gemsva(const unsigned n, float d, float* A,
		    float *x, float *y)
  { sspmv_(JOB_STR+5, &n, &d, A, x, &N_ONE, &S_ONE, y, &N_ONE); }
  inline void gemsva(const unsigned n, scomp d, scomp* A,
		    scomp *x, scomp *y)
  { chpmv_(JOB_STR+5, &n, &d, A, x, &N_ONE, &C_ONE, y, &N_ONE); }
  inline void gemsva(const unsigned n, dcomp d, dcomp* A,
		    dcomp *x, dcomp *y)
  { zhpmv_(JOB_STR+5, &n, &d, A, x, &N_ONE, &Z_ONE, y, &N_ONE); }
  
  
  // C = d A B, A is m x p, B is p x n
  inline void gemm(unsigned m, unsigned p, unsigned n, double d,
		   double* A, unsigned ldA, double* B, unsigned ldB,
		   double* C, unsigned ldC) {
    dgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &D_ZERO, C, &ldC);
  }
  inline void gemm(unsigned m, unsigned p, unsigned n, float d,
		   float* A, unsigned ldA, float* B, unsigned ldB,
		   float* C, unsigned ldC) {
    sgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &S_ZERO, C, &ldC);
  }
  inline void gemm(unsigned m, unsigned p, unsigned n, scomp d,
		   scomp* A, unsigned ldA, scomp* B, unsigned ldB,
		   scomp* C, unsigned ldC) {
    cgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &C_ZERO, C, &ldC);
  }
  inline void gemm(unsigned m, unsigned p, unsigned n, dcomp d,
		   dcomp* A, unsigned ldA, dcomp* B, unsigned ldB,
		   dcomp* C, unsigned ldC) {
    zgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &Z_ZERO, C, &ldC);
  }


  // C += d A B, A is m x p, B is p x n
  inline void gemma(unsigned m, unsigned p, unsigned n, double d,
		   double* A, unsigned ldA, double* B, unsigned ldB,
		   double* C, unsigned ldC) {
    dgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &D_ONE, C, &ldC);
  }
  inline void gemma(unsigned m, unsigned p, unsigned n, float d,
		   float* A, unsigned ldA, float* B, unsigned ldB,
		   float* C, unsigned ldC) {
    sgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &S_ONE, C, &ldC);
  }
  inline void gemma(unsigned m, unsigned p, unsigned n, scomp d,
		   scomp* A, unsigned ldA, scomp* B, unsigned ldB,
		   scomp* C, unsigned ldC) {
    cgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &C_ONE, C, &ldC);
  }
  inline void gemma(unsigned m, unsigned p, unsigned n, dcomp d,
		   dcomp* A, unsigned ldA, dcomp* B, unsigned ldB,
		   dcomp* C, unsigned ldC) {
    zgemm_(JOB_STR, JOB_STR, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &Z_ONE, C, &ldC);
  }

  // C = d A^T B, A is m x p, B is m x n
  inline void gemtm(unsigned m, unsigned p, unsigned n, double d,
		    double* A, unsigned ldA, double *B, unsigned ldB,
		    double* C, unsigned ldC) {
    dgemm_(JOB_STR+1, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &D_ZERO, C, &ldC);
  }
  inline void gemtm(unsigned m, unsigned p, unsigned n, float d,
		    float* A, unsigned ldA, float *B, unsigned ldB,
		    float* C, unsigned ldC) {
    sgemm_(JOB_STR+1, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &S_ZERO, C, &ldC);
  }
  inline void gemtm(unsigned m, unsigned p, unsigned n, scomp d,
		    scomp* A, unsigned ldA, scomp *B, unsigned ldB,
		    scomp* C, unsigned ldC) {
    cgemm_(JOB_STR+7, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &C_ZERO, C, &ldC);
  }
  inline void gemtm(unsigned m, unsigned p, unsigned n, dcomp d,
		    dcomp* A, unsigned ldA, dcomp *B, unsigned ldB,
		    dcomp* C, unsigned ldC) {
    zgemm_(JOB_STR+7, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &Z_ZERO, C, &ldC);
  }

  // C += d A^T B, A is m x p, B is m x n
  inline void gemtma(unsigned m, unsigned p, unsigned n, double d,
		    double* A, unsigned ldA, double *B, unsigned ldB,
		    double* C, unsigned ldC) {
    dgemm_(JOB_STR+1, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &D_ONE, C, &ldC);
  }
  inline void gemtma(unsigned m, unsigned p, unsigned n, float d,
		    float* A, unsigned ldA, float *B, unsigned ldB,
		    float* C, unsigned ldC) {
    sgemm_(JOB_STR+1, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &S_ONE, C, &ldC);
  }
  inline void gemtma(unsigned m, unsigned p, unsigned n, scomp d,
		    scomp* A, unsigned ldA, scomp *B, unsigned ldB,
		    scomp* C, unsigned ldC) {
    cgemm_(JOB_STR+7, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &C_ONE, C, &ldC);
  }
  inline void gemtma(unsigned m, unsigned p, unsigned n, dcomp d,
		    dcomp* A, unsigned ldA, dcomp *B, unsigned ldB,
		    dcomp* C, unsigned ldC) {
    zgemm_(JOB_STR+7, JOB_STR, &p, &n, &m, &d, A, &ldA, B, &ldB,
	   &Z_ONE, C, &ldC);
  }

  // C = d A B^T, A is m x p, B is n x p
  inline void gemmt(unsigned m, unsigned p, unsigned n, double d,
		    double* A, unsigned ldA, double *B, unsigned ldB,
		    double* C, unsigned ldC) {
    dgemm_(JOB_STR, JOB_STR+1, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &D_ZERO, C, &ldC);
  }
  inline void gemmt(unsigned m, unsigned p, unsigned n, float d,
		    float* A, unsigned ldA, float *B, unsigned ldB,
		    float* C, unsigned ldC) {
    sgemm_(JOB_STR, JOB_STR+1, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &S_ZERO, C, &ldC);
  }
  inline void gemmt(unsigned m, unsigned p, unsigned n, scomp d,
		    scomp* A, unsigned ldA, scomp *B, unsigned ldB,
		    scomp* C, unsigned ldC) {
    cgemm_(JOB_STR, JOB_STR+7, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &C_ZERO, C, &ldC);
  }
  inline void gemmt(unsigned m, unsigned p, unsigned n, dcomp d,
		    dcomp* A, unsigned ldA, dcomp *B, unsigned ldB,
		    dcomp* C, unsigned ldC) {
    zgemm_(JOB_STR, JOB_STR+7, &m, &n, &p, &d, A, &ldA, B, &ldB,
	   &Z_ZERO, C, &ldC);
  }
  
  // C += d A B^T, A is m x p, B is n x p
  inline void gemmta(unsigned m, unsigned p, unsigned n,
		    double* A, unsigned ldA, double *B, unsigned ldB,
		    double* C, unsigned ldC) {
    dgemm_(JOB_STR, JOB_STR+1, &m, &n, &p, &D_ONE, A, &ldA, B, &ldB,
	   &D_ONE, C, &ldC);
  }
  inline void gemmta(unsigned m, unsigned p, unsigned n,
		    float* A, unsigned ldA, float *B, unsigned ldB,
		    float* C, unsigned ldC) {
    sgemm_(JOB_STR, JOB_STR+1, &m, &n, &p, &S_ONE, A, &ldA, B, &ldB,
	   &S_ONE, C, &ldC);
  }
  inline void gemmta(unsigned m, unsigned p, unsigned n,
		    scomp* A, unsigned ldA, scomp *B, unsigned ldB,
		    scomp* C, unsigned ldC) {
    cgemm_(JOB_STR, JOB_STR+7, &m, &n, &p, &C_ONE, A, &ldA, B, &ldB,
	   &C_ONE, C, &ldC);
  }
  inline void gemmta(unsigned m, unsigned p, unsigned n,
		    dcomp* A, unsigned ldA, dcomp *B, unsigned ldB,
		    dcomp* C, unsigned ldC) {
    zgemm_(JOB_STR, JOB_STR+7, &m, &n, &p, &Z_ONE, A, &ldA, B, &ldB,
	   &Z_ONE, C, &ldC);
  }

  // C += d A^T A, C is a symm. matrix (packed upper half is stored), A is mxn
  inline void symtm(unsigned m, unsigned n, double* A, double d, double* C) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<=j; ++i) {
	double sum = 0.0;
	for (unsigned k=0; k<m; ++k) sum += A[k+i*m] * A[k+j*m];
	C[i+j*(j+1)/2] += d * sum;
      }
    }
  }
  inline void symtm(unsigned m, unsigned n, float* A, float d, float* C) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<=j; ++i) {
	float sum = 0.0;
	for (unsigned k=0; k<m; ++k) sum += A[k+i*m] * A[k+j*m];
	C[i+j*(j+1)/2] += d * sum;
      }
    }
  }
  inline void symtm(unsigned m, unsigned n, scomp* A, scomp d, scomp* C) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<=j; ++i) {
        scomp sum = C_ZERO;
        for (unsigned k=0; k<m; ++k) sum = sum + conj(A[k+i*m]) * A[k+j*m];
        C[i+j*(j+1)/2] = C[i+j*(j+1)/2] + d * sum;
      }
    }
  }
  inline void symtm(unsigned m, unsigned n, dcomp* A, dcomp d, dcomp* C) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<=j; ++i) {
        dcomp sum = Z_ZERO;
        for (unsigned k=0; k<m; ++k) sum = sum + conj(A[k+i*m]) * A[k+j*m];
        C[i+j*(j+1)/2] = C[i+j*(j+1)/2] + d * sum;
      }
    }
  }

  // C += d A A^T, C is a symm. matrix (packed upper half is stored), A is mxn
  inline void symmt(unsigned m, unsigned n, double* A, double d, double* C) {
    for (unsigned k=0; k<n; ++k) {
      for (unsigned j=0; j<=n; ++j) {
	double e = d * A[j+k*m];
	for (unsigned i=0; i<j; ++i) C[i+j*(j+1)/2] += e * A[i+k*m];
      }
    }
  }
  inline void symmt(unsigned m, unsigned n, float* A, float d, float* C) {
    for (unsigned k=0; k<n; ++k) {
      for (unsigned j=0; j<n; ++j) {
	float e = d * A[j+k*m];
	for (unsigned i=0; i<=j; ++i) C[i+j*(j+1)/2] += e * A[i+k*m];
      }
    }
  }
  inline void symmt(unsigned m, unsigned n, scomp* A, scomp d, scomp* C) {
    for (unsigned k=0; k<n; ++k) {
      for (unsigned j=0; j<n; ++j) {
        scomp e = d * conj(A[j+k*m]);
        for (unsigned i=0; i<=j; ++i)
	  C[i+j*(j+1)/2] = C[i+j*(j+1)/2] + e * A[i+k*m];
      }
    }
  }
  inline void symmt(unsigned m, unsigned n, dcomp* A, dcomp d, dcomp* C) {
    for (unsigned k=0; k<n; ++k) {
      for (unsigned j=0; j<n; ++j) {
        dcomp e = d * conj(A[j+k*m]);
        for (unsigned i=0; i<=j; ++i)
	  C[i+j*(j+1)/2] = C[i+j*(j+1)/2]+ e * A[i+k*m];
      }
    }
  }
  // Singular Value Decomposition
  inline int gesvd(unsigned m, unsigned n, double* A, double* S,
		   double* U, unsigned ldU, double* VT, unsigned ldVT,
		   unsigned nwk, double* wk) {
    int INF;
    dgesvd_(JOB_STR+2, JOB_STR+3, &m, &n, A, &m, S, U, &ldU, VT, &ldVT,
	    wk, &nwk, &INF);
    return INF;
  }
  inline int gesvd(unsigned m, unsigned n, float* A, float* S,
		   float* U, unsigned ldU, float* VT, unsigned ldVT,
		   unsigned nwk, float* wk) {
    int INF;
    sgesvd_(JOB_STR+2, JOB_STR+3, &m, &n, A, &m, S, U, &ldU, VT, &ldVT,
	    wk, &nwk, &INF);
    return INF;
  }
  inline int gesvd(unsigned m, unsigned n, scomp* A, float* S,
		   scomp* U, unsigned ldU, scomp* VT, unsigned ldVT,
		   unsigned nwk, scomp* wk) {
    int INF;
    float* rwk = new float[5*MIN(m,n)];
    cgesvd_(JOB_STR+2, JOB_STR+3, &m, &n, A, &m, S, U, &ldU, VT, &ldVT,
	    wk, &nwk, rwk, &INF);
    delete [] rwk;
    return INF;
  }
  inline int gesvd(unsigned m, unsigned n, dcomp* A, double* S,
		   dcomp* U, unsigned ldU, dcomp* VT, unsigned ldVT,
		   unsigned nwk, dcomp* wk) {
    int INF;
    double* rwk = new double[5*MIN(m,n)];
    zgesvd_(JOB_STR+2, JOB_STR+3, &m, &n, A, &m, S, U, &ldU, VT, &ldVT,
	    wk, &nwk, rwk, &INF);
    delete [] rwk;
    return INF;
  }
  // compute singular values
  inline int svals(unsigned m, unsigned n, double* A, double* S,
		   unsigned nwk, double* wk) {
    int INF;
    dgesvd_(JOB_STR, JOB_STR, &m, &n, A, &m, S, A, &m, A, &n, wk, &nwk, &INF);
    return INF;
  }
  inline int svals(unsigned m, unsigned n, float* A, float* S,
		   unsigned nwk, float* wk) {
    int INF;
    sgesvd_(JOB_STR, JOB_STR, &m, &n, A, &m, S, A, &m, A, &n, wk, &nwk, &INF);
    return INF;
  }
  inline int svals(unsigned m, unsigned n, scomp* A, float* S,
		   unsigned nwk, scomp* wk) {
    int INF;
    float *rwk = new float[5*MIN(m,n)];
    cgesvd_(JOB_STR, JOB_STR, &m, &n, A, &m, S, A, &m, A, &n,
	    wk, &nwk, rwk, &INF);
    delete [] rwk;
    return INF;
  }
  inline int svals(unsigned m, unsigned n, dcomp* A, double* S,
		   unsigned nwk, dcomp* wk) {
    int INF;
    double *rwk = new double[5*MIN(m,n)];
    zgesvd_(JOB_STR, JOB_STR, &m, &n, A, &m, S, A, &m, A, &n,
	    wk, &nwk, rwk, &INF);
    delete [] rwk;
    return INF;
  }
  // triangular factorisation
  inline int getrf(const unsigned n, double* A, unsigned* ipiv) {
    int INF;
    dgetrf_(&n, &n, A, &n, ipiv, &INF);
    return INF;
  }
  inline int getrf(const unsigned n, float* A, unsigned* ipiv) {
    int INF;
    sgetrf_(&n, &n, A, &n, ipiv, &INF);
    return INF;
  }
  inline int getrf(const unsigned n, scomp* A, unsigned* ipiv) {
    int INF;
    cgetrf_(&n, &n, A, &n, ipiv, &INF);
    return INF;
  }
  inline int getrf(const unsigned n, dcomp* A, unsigned* ipiv) {
    int INF;
    zgetrf_(&n, &n, A, &n, ipiv, &INF);
    return INF;
  }
  // packed triangular MV
  inline void tpmv(const unsigned n, double* A, double* x) {
    dtpmv_(JOB_STR+6, JOB_STR+1, JOB_STR+5, &n, A, x, &N_ONE);
  }
  inline void tpmv(const unsigned n, float* A, float* x) {
    stpmv_(JOB_STR+6, JOB_STR+1, JOB_STR+5, &n, A, x, &N_ONE);
  }
  inline void tpmv(const unsigned n, scomp* A, scomp* x) {
    ctpmv_(JOB_STR+6, JOB_STR+7, JOB_STR+5, &n, A, x, &N_ONE);
  }
  inline void tpmv(const unsigned n, dcomp* A, dcomp* x) {
    ztpmv_(JOB_STR+6, JOB_STR+7, JOB_STR+5, &n, A, x, &N_ONE);
  }
  // QR factorisation
  inline int geqrf(const unsigned m, const unsigned n, double* A,
		   double* tau, unsigned nwk, double* wk) {
    int INF;
    dgeqrf_(&m, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }
  inline int geqrf(const unsigned m, const unsigned n, float* A,
		   float* tau, unsigned nwk, float* wk) {
    int INF;
    sgeqrf_(&m, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }

  inline int geqrf(const unsigned m, const unsigned n, scomp* A,
		   scomp* tau, unsigned nwk, scomp* wk) {
    int INF;
    cgeqrf_(&m, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }
  
  inline int geqrf(const unsigned m, const unsigned n, dcomp* A,
		   dcomp* tau, unsigned nwk, dcomp* wk) {
    int INF;
    zgeqrf_(&m, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }

  // Multiply a general Matrix with the Q-Matrix (QR factorization)
  inline int ormqr(const unsigned m, const unsigned n, const unsigned p,
		   double* A, double* tau, double* C,
		   unsigned nwk, double* wk) {
    int INF;
    dormqr_(JOB_STR+6, JOB_STR, &m, &n, &p, A, &m, tau, C, &m, wk, &nwk, &INF);
    return INF;
  }
  inline int ormqr(const unsigned m, const unsigned n, const unsigned p,
		   float* A, float* tau, float* C,
		   unsigned nwk, float* wk) {
    int INF;
    sormqr_(JOB_STR+6, JOB_STR, &m, &n, &p, A, &m, tau, C, &m, wk, &nwk, &INF);
    return INF;
  }
  inline int ormqr(const unsigned m, const unsigned n, const unsigned p,
		   scomp* A, scomp* tau, scomp* C,
		   unsigned nwk, scomp* wk) {
    int INF;
    cunmqr_(JOB_STR+6, JOB_STR, &m, &n, &p, A, &m, tau, C, &m, wk, &nwk, &INF);
    return INF;
  }
  inline int ormqr(const unsigned m, const unsigned n, const unsigned p,
		   dcomp* A, dcomp* tau, dcomp* C,
		   unsigned nwk, dcomp* wk) {
    int INF;
    zunmqr_(JOB_STR+6, JOB_STR, &m, &n, &p, A, &m, tau, C, &m, wk, &nwk, &INF);
    return INF;
  }

  //return Q-Matrix (QR factorization) in A
  inline int orgqr(const unsigned m, const unsigned n, double* A, double* tau,
		   unsigned nwk, double* wk) {
    int INF;
    dorgqr_(&m, &n, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }
  inline int orgqr(const unsigned m, const unsigned n, float* A, float* tau,
		   unsigned nwk, float* wk) {
    int INF;
    sorgqr_(&m, &n, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }
  inline int orgqr(const unsigned m, const unsigned n, scomp* A, scomp* tau,
		   unsigned nwk, scomp* wk) {
    int INF;
    cungqr_(&m, &n, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }
  inline int orgqr(const unsigned m, const unsigned n, dcomp* A, dcomp* tau,
		   unsigned nwk, dcomp* wk) {
    int INF;
    zungqr_(&m, &n, &n, A, &m, tau, wk, &nwk, &INF);
    return INF;
  }
  
  // B=A^T  A in R^(m,n)
  inline void transpose(unsigned m, unsigned n, double* A, double* B) {
    for (unsigned i=0; i<m; ++i) dcopy_(&n, A+i, &m, B+i*n, &N_ONE);
  }
  inline void transpose(unsigned m, unsigned n, float* A, float* B) {
    for (unsigned i=0; i<m; ++i) scopy_(&n, A+i, &m, B+i*n, &N_ONE);
  }
  inline void transpose(unsigned m, unsigned n, scomp* A, scomp* B) {
    for(unsigned i=0; i<m; i++)
      for(unsigned j=0; j<n; j++) B[j+i*n] = conj(A[i+j*m]);
  }
  inline void transpose(unsigned m, unsigned n, dcomp* A, dcomp* B) {
    for(unsigned i=0; i<m; i++)
      for(unsigned j=0; j<n; j++) B[j+i*n] = conj(A[i+j*m]);
  }

  // product of two upper triangular matrices M = R1 R2^T, R1 mxp, R2 nxp
  inline void utrmmt(unsigned m, unsigned p, unsigned n, double* R1,
		     unsigned ldR1, double* R2, unsigned ldR2, double* M) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<m; ++i) {
	double d = D_ZERO;
	for (unsigned l=MAX(i,j); l<p; ++l) d += R1[i+l*ldR1] * R2[j+l*ldR2];
	M[i+j*m] = d;
      }
    }
  }
  inline void utrmmt(unsigned m, unsigned p, unsigned n, float* R1,
		     unsigned ldR1, float* R2, unsigned ldR2, float* M) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<m; ++i) {
	float d = S_ZERO;
	for (unsigned l=MAX(i,j); l<p; ++l) d += R1[i+l*ldR1] * R2[j+l*ldR2];
	M[i+j*m] = d;
      }
    }
  }
  inline void utrmmt(unsigned m, unsigned p, unsigned n, scomp* R1,
		     unsigned ldR1, scomp* R2, unsigned ldR2, scomp* M) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<m; ++i) {
	scomp d = C_ZERO;
        for (unsigned l=MAX(i,j); l<p; ++l)
	  d = d + R1[i+l*ldR1] * conj(R2[j+l*ldR2]);
	M[i+j*m] = d;
      }
    }
  }
  inline void utrmmt(unsigned m, unsigned p, unsigned n, dcomp* R1,
		     unsigned ldR1, dcomp* R2, unsigned ldR2, dcomp* M) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<m; ++i) {
	dcomp d = Z_ZERO;
        for (unsigned l=MAX(i,j); l<p; ++l)
	  d = d + R1[i+l*ldR1] * conj(R2[j+l*ldR2]);
	M[i+j*m] = d;
      }
    }
  }
  
  inline void fill0_ltr(unsigned n, double* A) {
    for (unsigned j=0; j<n; ++j) {
      *A++ = (double) j;
      for (unsigned i=j+1; i<n; ++i) *A++ = D_ZERO;
    }
  }
  inline void fill0_ltr(unsigned n, float* A) {
    for (unsigned j=0; j<n; ++j) {
      *A++ = (float) j;
      for (unsigned i=j+1; i<n; ++i) *A++ = S_ZERO;
    }
  }
  inline void fill0_ltr(unsigned n, scomp* A) {
    for (unsigned j=0; j<n; ++j) {
      A->re = (float) j; A->im = S_ZERO;
      ++A;
      for (unsigned i=j+1; i<n; ++i) *A++ = C_ZERO;
    }
  }
  inline void fill0_ltr(unsigned n, dcomp* A) {
    for (unsigned j=0; j<n; ++j) {
      A->re = (double) j; A->im = D_ZERO;
      ++A;
      for (unsigned i=j+1; i<n; ++i) *A++ = Z_ZERO;
    }
  }
  
  // fill nxn matrix A with identity
  inline void fillId(unsigned n, double *A) {
    for (unsigned j=0; j<n; ++j) {
      unsigned i = 0;
      for (; i<j; ++i) *A++ = D_ZERO;
      *A++ = D_ONE; ++i;
      for (; i<n; ++i) *A++ = D_ZERO;
    }
  }  
  inline void fillId(unsigned n, float *A) {
    for (unsigned j=0; j<n; ++j) {
      unsigned i = 0;
      for (; i<j; ++i) *A++ = S_ZERO;
      *A++ = S_ONE; ++i;
      for (; i<n; ++i) *A++ = S_ZERO;
    }
  }
  inline void fillId(unsigned n, scomp *A) {
    for (unsigned j=0; j<n; ++j) {
      unsigned i = 0;
      for (; i<j; ++i) *A++ = C_ZERO;
      *A++ = C_ONE; ++i;
      for (; i<n; ++i) *A++ = C_ZERO;
    }
  }
  inline void fillId(unsigned n, dcomp *A) {
    for (unsigned j=0; j<n; ++j) {
      unsigned i = 0;
      for (; i<j; ++i) *A++ = Z_ZERO;
      *A++ = Z_ONE; ++i;
      for (; i<n; ++i) *A++ = Z_ZERO;
    }
  }

  // fill nxn upper triang. matrix A with identity (packed storage)
  inline void fillId_utr(unsigned n, double *A) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<j; ++i) *A++ = D_ZERO;
      *A++ = D_ONE;
    }
  }
  inline void fillId_utr(unsigned n, float *A) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<j; ++i) *A++ = S_ZERO;
      *A++ = S_ONE;
    }
  }
  inline void fillId_utr(unsigned n, scomp *A) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<j; ++i) *A++ = C_ZERO;
      *A++ = C_ONE;
    }
  }
  inline void fillId_utr(unsigned n, dcomp *A) {
    for (unsigned j=0; j<n; ++j) {
      for (unsigned i=0; i<j; ++i) *A++ = Z_ZERO;
      *A++ = Z_ONE;
    }
  }

  // fill nxn normalized lower triang. matrix A with identity (packed storage)
  inline void fillId_ltr(unsigned n, double *A) {
    for (unsigned i=0; i<n; ++i) {
      for (unsigned j=0; j<i; ++j) *A++ = D_ZERO;
      // for pivoting, a ltr is assumed to have ones on the diagonal
      *A++ = (double) i;
    }
  }
  inline void fillId_ltr(unsigned n, float *A) {
    for (unsigned i=0; i<n; ++i) {
      for (unsigned j=0; j<i; ++j) *A++ = S_ZERO;
      // for pivoting, a ltr is assumed to have ones on the diagonal
      *A++ = (float) i;  
    }
  }
  inline void fillId_ltr(unsigned n, scomp *A) {
    for (unsigned i=0; i<n; ++i) {
      for (unsigned j=0; j<i; ++j) *A++ = C_ZERO;
      // for pivoting, a ltr is assumed to have ones on the diagonal
      A->re = (float) i; A->im = S_ZERO;
      ++A;
    }
  }
  inline void fillId_ltr(unsigned n, dcomp *A) {
    for (unsigned i=0; i<n; ++i) {
      for (unsigned j=0; j<i; ++j) *A++ = Z_ZERO;
      // for pivoting, a ltr is assumed to have ones on the diagonal
      A->re = (double) i; A->im = D_ZERO;
      ++A;
    }
  }
}


namespace lapack {

  // general inversion
  inline void geinv(const unsigned n, double* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    double* const work = new double[lwork];
    assert(work!=NULL);

    int INFO = blas::getrf(n, A, ipiv);
    assert(INFO==0);

    dgetri_(&n, A, &n, ipiv, work, &lwork, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  inline void geinv(const unsigned n, float* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    float* const work = new float[lwork];
    assert(work!=NULL);

    int INFO = blas::getrf(n, A, ipiv);
    assert(INFO==0);

    sgetri_(&n, A, &n, ipiv, work, &lwork, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  inline void geinv(const unsigned n, scomp* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    scomp* const work = new scomp[lwork];
    assert(work!=NULL);

    int INFO = blas::getrf(n, A, ipiv);
    assert(INFO==0);

    cgetri_(&n, A, &n, ipiv, work, &lwork, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  inline void geinv(const unsigned n, dcomp* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    dcomp* const work = new dcomp[lwork];
    assert(work!=NULL);

    int INFO = blas::getrf(n, A, ipiv);
    assert(INFO==0);

    zgetri_(&n, A, &n, ipiv, work, &lwork, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  
  inline void geinv_sym(const unsigned n, double* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    double* const work = new double[lwork];
    assert(work!=NULL);

    int INFO;
    dsptrf_(JOB_STR+5, &n, A, ipiv, &INFO);
    assert(INFO==0);

    dsptri_(JOB_STR+5, &n, A, ipiv, work, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  inline void geinv_sym(const unsigned n, float* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    float* const work = new float[lwork];
    assert(work!=NULL);

    int INFO;
    ssptrf_(JOB_STR+5, &n, A, ipiv, &INFO);
    assert(INFO==0);

    ssptri_(JOB_STR+5, &n, A, ipiv, work, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  inline void geinv_sym(const unsigned n, scomp* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    scomp* const work = new scomp[lwork];
    assert(work!=NULL);

    int INFO;
    csptrf_(JOB_STR+5, &n, A, ipiv, &INFO);
    assert(INFO==0);

    csptri_(JOB_STR+5, &n, A, ipiv, work, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }
  inline void geinv_sym(const unsigned n, dcomp* const A) {
    unsigned* const ipiv = new unsigned[n];
    assert(ipiv!=NULL);

    const unsigned lwork = 4*n;
    dcomp* const work = new dcomp[lwork];
    assert(work!=NULL);

    int INFO;
    zsptrf_(JOB_STR+5, &n, A, ipiv, &INFO);
    assert(INFO==0);

    zsptri_(JOB_STR+5, &n, A, ipiv, work, &INFO);
    assert(INFO==0);

    delete [] work;
    delete [] ipiv;
  }

  // packed triangular factorisation
  inline int pptrf(const unsigned n, double* A) {
    int inf;
	dpptrf_(JOB_STR+5, &n, A, &inf);
	return inf;
  }
  inline int pptrf(const unsigned n, float* A) {
    int inf;
    spptrf_(JOB_STR+5, &n, A, &inf);
    return inf;
  }
  inline int pptrf(const unsigned n, scomp* A) {
    int inf;
    cpptrf_(JOB_STR+5, &n, A, &inf);
    return inf;
  }
  inline int pptrf(const unsigned n, dcomp* A) {
    int inf;
    zpptrf_(JOB_STR+5, &n, A, &inf);
    return inf;
  }
  // lower triangular solve
  inline int ltrs(const unsigned n, double* A,
		  const unsigned p, double* B, const unsigned ldB) {
    int inf;
    dtptrs_(JOB_STR+6, JOB_STR, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int ltrs(const unsigned n, float* A,
		  const unsigned p, float* B, const unsigned ldB) {
    int inf;
    stptrs_(JOB_STR+6, JOB_STR, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int ltrs(const unsigned n, scomp* A,
		  const unsigned p, scomp* B, const unsigned ldB) {
    int inf;
    ctptrs_(JOB_STR+6, JOB_STR, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
    inline int ltrs(const unsigned n, dcomp* A,
		  const unsigned p, dcomp* B, const unsigned ldB) {
    int inf;
    ztptrs_(JOB_STR+6, JOB_STR, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  // lower triangular solve
  inline int ltrTs(const unsigned n, double* A,
		   const unsigned p, double* B, const unsigned ldB) {
    int inf;
    dtptrs_(JOB_STR+6, JOB_STR+1, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int ltrTs(const unsigned n, float* A,
		   const unsigned p, float* B, const unsigned ldB) {
    int inf;
    stptrs_(JOB_STR+6, JOB_STR+1, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int ltrTs(const unsigned n, scomp* A,
		   const unsigned p, scomp* B, const unsigned ldB) {
    int inf;
    ctptrs_(JOB_STR+6, JOB_STR+7, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int ltrTs(const unsigned n, dcomp* A,
		   const unsigned p, dcomp* B, const unsigned ldB) {
    int inf;
    ztptrs_(JOB_STR+6, JOB_STR+7, JOB_STR+5, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  // upper triangular solve
  inline int utrs(const unsigned n, double* A,
		  const unsigned p, double* B, const unsigned ldB) {
    int inf;
    dtptrs_(JOB_STR+5, JOB_STR, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int utrs(const unsigned n, float* A,
		  const unsigned p, float* B, const unsigned ldB) {
    int inf;
    stptrs_(JOB_STR+5, JOB_STR, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int utrs(const unsigned n, scomp* A,
		  const unsigned p, scomp* B, const unsigned ldB) {
    int inf;
    ctptrs_(JOB_STR+5, JOB_STR, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int utrs(const unsigned n, dcomp* A,
		  const unsigned p, dcomp* B, const unsigned ldB) {
    int inf;
    ztptrs_(JOB_STR+5, JOB_STR, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }

  // upper triangluar transpose solve
  inline int utrTs(const unsigned n, double* A,
		   const unsigned p, double* B, const unsigned ldB) {
    int inf;
    dtptrs_(JOB_STR+5, JOB_STR+1, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int utrTs(const unsigned n, float* A,
		   const unsigned p, float* B, const unsigned ldB) {
    int inf;
    stptrs_(JOB_STR+5, JOB_STR+1, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int utrTs(const unsigned n, scomp* A,
		   const unsigned p, scomp* B, const unsigned ldB) {
    int inf;
    ctptrs_(JOB_STR+5, JOB_STR+7, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
  inline int utrTs(const unsigned n, dcomp* A,
		   const unsigned p, dcomp* B, const unsigned ldB) {
    int inf;
    ztptrs_(JOB_STR+5, JOB_STR+7, JOB_STR, &n, &p, A, B, &ldB, &inf);
    return inf;
  }
}

#endif
