#ifndef CLUSTERBBX_H
#define CLUSTERBBX_H

#include "cluster.h"

/*!
  \brief A class for storing clusters of degrees of freedom.
  Subdivision is based on bounding boxes (bbx)
*/
class cluster_bbx : public cluster
{
  double cntrdir;           // the center with respect to the main direction
  double *xminmax;
  unsigned maindir;         // main direction

  double dist2(cluster_bbx*) const;

 public:
  bool isadm(double eta, cluster* cl) const;
  void init(dof**);
  void subdivide(dof** dofs, unsigned bmin, unsigned& nClusters);
  virtual cluster* clone(dof**, unsigned, unsigned) const = 0;
  cluster_bbx(unsigned k, unsigned l) : cluster(k, l) { }
  virtual ~cluster_bbx() { delete [] xminmax; }
};

//! a class for storing clusters of degrees of freedom in 1d
class cluster1d_bbx : public cluster_bbx
{
 protected:
  unsigned dim() const { return 1; }
  cluster* clone(dof** dof, unsigned beg, unsigned end) const
  { return new cluster1d_bbx(dof, beg, end); }

 public:
  cluster1d_bbx(dof** dof, unsigned k, unsigned l) :
    cluster_bbx(k, l) { init(dof); }
};

//! a class for storing clusters of degrees of freedom in 2d
class cluster2d_bbx : public cluster_bbx
{
 protected:
  unsigned dim() const { return 2; }
  cluster* clone(dof** dof, unsigned beg, unsigned end) const
  { return new cluster2d_bbx(dof, beg, end); }

 public:
  cluster2d_bbx(dof** dof, unsigned k, unsigned l) :
    cluster_bbx(k, l) { init(dof); }
};


//! a class for storing clusters of degrees of freedom in 3d
class cluster3d_bbx : public cluster_bbx
{
 protected:
  unsigned dim() const { return 3; }
  cluster* clone(dof** dof, unsigned beg, unsigned end) const
  { return new cluster3d_bbx(dof, beg, end); }

 public:
  cluster3d_bbx(dof** dof, unsigned k, unsigned l) : cluster_bbx(k, l)
  { init(dof); }
};

#endif
