/*! \file  cluster_pca.h
    \brief Include file for the class cluster_pca
    \author M. Bebendorf
    \date   7/14/2004
*/

#ifndef CLUSTERPCA_H
#define CLUSTERPCA_H

#include "cluster.h"

/*!
  \brief A class for storing clusters of degrees of freedom.
  Subdivision is based on the principal component analysis (PCA)
*/
class cluster_pca : public cluster
{
  double* com;

  double dist2(cluster_pca*) const;

 public:
  bool isadm(double eta, cluster* cl) const;
  void init(dof**);
  void subdivide(dof** dofs, unsigned bmin, unsigned& nClusters);
  void subdivideND(dof** dofs, unsigned bmin, unsigned& nClusters);

  cluster_pca(unsigned k, unsigned l) : cluster(k, l) { }

  virtual cluster* clone(dof**, unsigned, unsigned) const = 0;
  virtual ~cluster_pca() { delete [] com; }
};


//! a class for storing clusters of degrees of freedom in 2d
class cluster2d_pca : public cluster_pca
{
 protected:
  unsigned dim() const { return 2; }
  cluster* clone(dof** dof, unsigned beg, unsigned end) const
  { return new cluster2d_pca(dof, beg, end); }

 public:
  cluster2d_pca(dof** dofs, unsigned k, unsigned l) :
    cluster_pca(k, l) { init(dofs); }
};


//! a class for storing clusters of degrees of freedom in 3d
class cluster3d_pca : public cluster_pca
{
 protected:
  unsigned dim() const { return 3; }
  cluster* clone(dof** dofs, unsigned beg, unsigned end) const
  { return new cluster3d_pca(dofs, beg, end); }

 public:
  cluster3d_pca(dof** dofs, unsigned k, unsigned l) :
    cluster_pca(k, l) { init(dofs); }
};


#endif
