/*! \file  dof.h
    \brief Include file for the base class dof
    \author M. Bebendorf
    \date   7/14/2004
*/

#ifndef DOF_H
#define DOF_H

#include <iostream>
#include "sll.h"

/*! \brief Base class for finite elements.
    An array of pointers of type dof is needed for clustering.
    This array will be reordered to assure that geometrically close dofs 
    have a close index (vidx).
*/
class dof
{
 public:
  /*!
    \brief the index of the associated variable (changed during clustering),
           needs to be initialized together with idx
           
    \verbatim
    for (i=0; i<nPanels; ++i) {
      clindxPan[i] = &Panels[i];
      Panels[i].idx = Panels[i].vidx = i;
    }
    \endverbatim
  */
  unsigned vidx;

  //! index of the dof (remains unchanged during clustering)
  unsigned idx;

  //! the dimension of the Euklidiean space the finite element lives in
  virtual unsigned getdim() const = 0;

  //! returns i-th coordinate of the center
  virtual double getcenter(const unsigned) const = 0;

  //! returns the squared radius of an enclosing ball
  virtual double getradius2() const { return 0.0; }

  virtual void getneighbors(sll<dof*>&) const { }
};

#endif
