#ifndef GAUSS2D_H
#define GAUSS2D_H

#include "vec3d.h"
extern void gauss2d(unsigned i, unsigned& ng, double*& gp, double*& gw);


// ****************************************************************************
// 7 Punkte Gauss, sechster Ordnung

inline void gauss_weights_7pt(unsigned& ng, double*& gw)
// liefert Standard-Integrationsgewichte fuer 7 Punkte Gauss-Quadratur
// (d.h. ohne Skalierung mit Flaecheninhalt) 
{
  static double Gauss7pt[] = {
    0.225, 0.12593918054482715260, 0.12593918054482715260,
    0.12593918054482715260, 0.13239415278850618074, 0.13239415278850618074,
    0.13239415278850618074
  };

  ng = 7;
  gw = Gauss7pt;
}


inline void gauss_nodes_7pt(vec3d& v1, vec3d& v2, vec3d& v3, vec3d int_node[7])
// Bestimmt die Integrationsknoten im betrachteten Dreieck. 
{
  static double gp[] = {
    0.33333333333333333333,
    0.10128650732345633880,
    0.79742698535308732240,
    0.47014206410511508977,
    0.05971587178976982046
  };

  int_node[0] = (v1 + v2 + v3) * gp[0];
  int_node[1] = (v2 + v3)*gp[1] + v1*gp[2];
  int_node[2] = (v1 + v3)*gp[1] + v2*gp[2];
  int_node[3] = (v1 + v2)*gp[1] + v3*gp[2];
  int_node[4] = (v2 + v3)*gp[3] + v1*gp[4];
  int_node[5] = (v1 + v2)*gp[3] + v3*gp[4];
  int_node[6] = (v1 + v3)*gp[3] + v2*gp[4];
}



#endif
