#ifndef VEC2_H
#define VEC2_H

#include <math.h>

class vec2 {

  double x[2];

 public:
  vec2() {};
  vec2(const double d1, const double d2) {
    x[0] = d1;
    x[1] = d2;
  };

  ~vec2() {};

  double operator[](const unsigned i) const { return x[i]; }

  double sqrsum() const { return x[0]*x[0] + x[1]*x[1]; };
  double norm2() const { return sqrt(sqrsum()); };

  vec2 rotate(const double a) const {
    return vec2(cos(a)*x[0]-sin(a)*x[1], sin(a)*x[0]+cos(a)*x[1]);
  }

  vec2 operator+ (const vec2& v) const {
    return vec2(x[0]+v.x[0], x[1]+v.x[1]);
  }

  vec2& operator+= (vec2& v) {
    x[0] += v.x[0];
    x[1] += v.x[1];
    return *this;
  }

  vec2 operator- (const vec2& v) const {
    return vec2(x[0]-v.x[0], x[1]-v.x[1]);
  }  

  vec2& operator-= (const vec2& v) {
    x[0] -= v.x[0];
    x[1] -= v.x[1];
    return *this;
  }

  vec2 operator* (const double a) const {
    return vec2(a*x[0], a*x[1]);
  }
  
  double operator* (const vec2& v) const {
    return x[0]*v.x[0]+x[1]*v.x[1];
  }

  vec2& operator*= (const double a) {
    x[0] *= a;
    x[1] *= a;
    return *this;
  }
};

#endif
