#include "femlib.h"
#ifndef BLF_h__
#define BLF_h__
/********************************************************************
	created:	2008/09/02
	created:	2:9:2008   9:13
	file base:	BLF
	file ext:	h
	author:		Florian Krmer
				Header file for the bilinearforms
				- <gradhat,gradhat>
				- <Edge,gradHat>
				- edge
				- <Edge, Edge>
				- <curlEdge, curlEdge>
				- N4
	purpose:	
*********************************************************************/
namespace femlib
{
	//////////////////////////////////////////////////////////////////////////
	///Bilinear Forms (Grad)
	//////////////////////////////////////////////////////////////////////////
	
	//************************************
	// Method:    get_BLF_gradHat_gradHat4
	// FullName:  femlib::get_BLF_gradHat_gradHat4
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Evaluates the BLF <grad N4_i, grad N4_j>
	//				The results are stored like:
	//								for( node_i = 0;  node_i <4;  node_i++ )
	//								{
	//									for( node_j = 0;  node_j < 4;  node_j++ )	
	//									{
	//										results[node_i*4+node_j] += (QUOT*gradN4[node_i]*gradN4[node_j]);
	//									}
	//								}
	// Parameter: double * GEOM
	// Parameter: double * results			size: 4*4
	//************************************
	_EXTERN_ void get_BLF_gradHat_gradHat4(double* GEOM, double *results );
	//************************************
	// Method:    get_BLF_gradHat_gradHat4_flat
	// FullName:  femlib::get_BLF_gradHat_gradHat4
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Evaluates the BLF <grad N4_i, grad N4_j>
	//				The results are stored like:
	//								for( node_i = 0;  node_i <4;  node_i++ )
	//								{
	//									for( node_j = 0;  node_j < 4;  node_j++ )	
	//									{
	//										results[node_i*4+node_j] += (QUOT*gradN4[node_i]*gradN4[node_j]);
	//									}
	//								}
	// Parameter: double * GEOM1			std. TET4 GEOM Array
	// Parameter: double * results			size: 4*4
	//************************************
	_EXTERN_ void get_BLF_gradHat_gradHat4_flat(double* GEOM, double *results );

	//************************************
	// Method:    gradHat_gradHat
	// FullName:  femlib::gradHat_gradHat
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: Evaluates the interaction between node_i and node_j for the in GEOM descripted TET10
	// Parameter: const unsigned i			//between 1-10
	// Parameter: const unsigned j			//between 1-10
	// Parameter: double * GEOM				//std. GEOM
	//************************************
	_EXTERN_ double gradHat_gradHat( const unsigned i, const unsigned j, double* GEOM );
	//////////////////////////////////////////////////////////////////////////
	//Edge Grad
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    get_BLF_Edge_gradN4
	// FullName:  femlib::get_BLF_Edge_gradN4
	// Access:    public 
	// Returns:   _EXTERN_ double			Result
	// Description: Returns the blf <Edge,GradN> between 
	//				localedge "edge1" and localnode "gradHat_Nr" for the curved TET4
	//				which is descripted by the std. GEOM array. the node_labels have also
	// Parameter: unsigned edge1			between 1-6
	// Parameter: unsigned gradHat_Nr		between 1-4
	// Parameter: double * GEOM				std. GEOM Array
	// Parameter: unsigned * node_labels	size: 10, filled with node_labels
	//************************************
	_EXTERN_ double get_BLF_Edge_gradN4(unsigned edge1, unsigned gradHat_Nr, double* GEOM, unsigned* node_labels);
	//************************************
	// Method:    get_BLF_Edge_gradN4_flat
	// FullName:  femlib::get_BLF_Edge_gradN4
	// Access:    public 
	// Returns:   _EXTERN_ double			Result
	// Description: Returns the blf <Edge,GradN> between 
	//				localedge "edge1" and localnode "gradHat_Nr" for the curved TET4
	//				which is descripted by the std. GEOM array. the node_labels have also
	// Parameter: unsigned edge1			between 1-6
	// Parameter: unsigned gradHat_Nr		between 1-4
	// Parameter: double * GEOM				std. GEOM Array (TET4)
	// Parameter: unsigned * node_labels	size: node_labels
	//************************************
	_EXTERN_ double get_BLF_Edge_gradN4_flat(unsigned edge1, unsigned gradHat_Nr, double* GEOM, unsigned* node_labels);

	//////////////////////////////////////////////////////////////////////////
	//Edge
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    get_BLF_1_Edge
	// FullName:  femlib::get_BLF_1_Edge
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: 
	// Parameter: unsigned edge1
	// Parameter: unsigned edge2
	// Parameter: double * GEOM
	// Parameter: unsigned * node_labels
	//************************************
	_EXTERN_ double get_BLF_1_Edge(unsigned edge1, unsigned edge2, double* GEOM, unsigned* node_labels);
	//////////////////////////////////////////////////////////////////////////
	//EdgeEdge
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    get_BLF_Edge_Edge
	// FullName:  femlib::get_BLF_Edge_Edge
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: Calculates the <EdgeFunction,Edgefunction'> BLF for an cuved TET4
	// Parameter: unsigned edge1
	// Parameter: unsigned edge2
	// Parameter: double * GEOM				std. TET10 GEOM Array
	// Parameter: unsigned * node_labels	
	//************************************
	_EXTERN_ double get_BLF_Edge_Edge(unsigned edge1, unsigned edge2, double* GEOM, unsigned* node_labels);
		
	//************************************
	// Method:    get_BLF_Edge_Edge
	// FullName:  femlib::get_BLF_Edge_Edge
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: Calculates the <EdgeFunction,Edgefunction'> BLF for an flat TET4
	// Parameter: unsigned edge1
	// Parameter: unsigned edge2
	// Parameter: double * GEOM			std. TET4 GEOM Array
	// Parameter: unsigned * node_labels
	//************************************
	_EXTERN_ double get_BLF_Edge_Edge_flat(unsigned edge1, unsigned edge2, double* GEOM, unsigned* node_labels);



	
	//////////////////////////////////////////////////////////////////////////
	// CURLS
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    get_BLF_curlEdge_curlEdge4
	// FullName:  femlib::get_BLF_curlEdge_curlEdge4
	// Access:    public 
	// Returns:   _EXTERN_ double			det
	// Description: Evaluates the BLF <curlEdge_1, curlEdge_2> for an TET
	// Parameter: unsigned edge1			first local edgenr
	// Parameter: unsigned edge2			second local edgenr
	// Parameter: unsigned * node_labels	list of the 10 nodelabels
	// Parameter: double * GEOM				std. GEOM
	//************************************
	_EXTERN_ double get_BLF_curlEdge_curlEdge4(unsigned edge1, unsigned edge2, unsigned* node_labels,double* GEOM);

	//************************************
	// Method:    get_BLF_curlEdge_curlEdge4
	// FullName:  femlib::get_BLF_curlEdge_curlEdge4
	// Access:    public 
	// Returns:   _EXTERN_ double			det
	// Description: Evaluates the BLF <curlEdge_1, curlEdge_2> for an TET4
	// Parameter: unsigned edge1			first local edgenr
	// Parameter: unsigned edge2			second local edgenr
	// Parameter: unsigned * node_labels	list of the 10 nodelabels
	// Parameter: double * GEOM				std. GEOM (TET4)	
	//************************************
	_EXTERN_ double get_BLF_curlEdge_curlEdge4_flat(unsigned edge1, unsigned edge2, unsigned* node_labels,double* GEOM);

	//////////////////////////////////////////////////////////////////////////
	// Hatfunction
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    get_BLF_N4
	// FullName:  femlib::get_BLF_N4
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: Evaluates int(N4_i)
	// Parameter: unsigned i		which nodal shape function
	//************************************
	_EXTERN_ double get_BLF_N4(unsigned i);

} // end femlib
#endif // BLF_h__
