
#include "vector2d.h"
#include "vector3d.h"
#include "defines.h"

#ifndef GeometryInformations_h__
#define GeometryInformations_h__



using namespace vector2d;
using namespace vector3d;
namespace femlib
{
	//////////////////////////////////////////////////////////////////////////
	//Nodal related
	//////////////////////////////////////////////////////////////////////////
	//! Get Coord. of Node 1,..,10 in local coord (xi,eta,zeta)
	//! These are the local coordinates of the Nodes
	//************************************
	// Method:    get_local_coord_nodal
	// FullName:  femlib::get_local_coord_nodal
	// Access:    public 
	// Returns:   _EXTERN_ Vec3
	// Description: Get the (xi,eta,zeta) coordinate of the nodes for the tet10
	// Parameter: unsigned i
	//************************************
	_EXTERN_ Vec3 get_local_coord_nodal(unsigned i);
	//! Get Coord. of Node 1,..,6 in local coord (xi,eta,zeta)
	//! These are the local coordinates of the Nodes of the tria
	//************************************
	// Method:    get_local_coord_nodal_2d
	// FullName:  femlib::get_local_coord_nodal_2d
	// Access:    public 
	// Returns:   _EXTERN_ Vec2
	// Description: Get the local coordinate of the tria6
	// Parameter: unsigned i
	//************************************
	_EXTERN_ Vec2 get_local_coord_nodal_2d(unsigned i);
	//////////////////////////////////////////////////////////////////////////
	//Edge related
	//////////////////////////////////////////////////////////////////////////
	//************************************
	// Method:    getEdgeNodes
	// FullName:  femlib::getEdgeNodes
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Get nodes that belong to an edge (node index starts with 1)
	// Parameter: unsigned edge_nr		Input: for which edge do you want to know the nodes (local numbering)
	// Parameter: unsigned & node_1		Output1: node1 to the edge edge_nr (local numbering)
	// Parameter: unsigned & node_2		Output2: node2 to the edge edge_nr (local numbering)
	//************************************
	_EXTERN_ void getEdgeNodes(unsigned edge_nr, unsigned &node_1, unsigned &node_2);
	//************************************
	// Method:    getEdgeNodes
	// FullName:  femlib::getEdgeNodes
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Get nodes that belong to an edge (node index starts with 0)
	// Parameter: unsigned edge_nr		Input: for which edge do you want to know the nodes (local numbering)
	// Parameter: unsigned & node_1		Output1: node1 to the edge edge_nr (local numbering)
	// Parameter: unsigned & node_2		Output2: node2 to the edge edge_nr (local numbering)
	//************************************
	_EXTERN_ void getEdgeNodes0(unsigned edge_nr, unsigned &node_1, unsigned &node_2);

	//!This methods defines the orientation of an edge
	//************************************
	// Method:    get_Signs_array
	// FullName:  femlib::get_Signs_array
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Calculates the signs for the edge function from the global nodelabels, which are provided by the node-labels array
	//				This function is called by get_edge function
	// Parameter: unsigned * node_labels	size: 4		The four nodelabels of the mainnodes (1,3,5,10)
	// Parameter: double * signs			size: 6		The signs for the edge functions
	//************************************
	_EXTERN_ void get_Signs_array(unsigned* node_labels, double* signs);
}



#endif // GeometryInformations_h__
