#ifndef GradFunctions_h__
#define GradFunctions_h__
namespace femlib
{
	///Evalute function
	//************************************
	// Method:    gradFunc
	// FullName:  femlib::gradFunc
	// Access:    public 
	// Returns:   _EXTERN_ Vec3
	// Description: Evaluates the grad(Function defined by arguments at xi,eta,zeta)
	// Parameter: double * GEOM			std. GEOM Array
	// Parameter: const double Xi		local coodinate 1
	// Parameter: const double Eta		local coodinate 2
	// Parameter: const double Zeta		local coodinate 3
	// Parameter: double * Arg			Arguments for gradHat function (for example potential)
	//************************************
	_EXTERN_ Vec3 gradFunc(double* GEOM,const double Xi, const double Eta,const double Zeta, double* Arg);
	//GRAD
	//************************************
	// Method:    get_Grad_N4
	// FullName:  femlib::get_Grad_N4
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: get Grad of the interpolated shapefunctions of the N4 routine. 
	// Parameter: double xi
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM
	// Parameter: Vec3 * gder4
	//************************************
	_EXTERN_ double get_Grad_N4(double xi, double eta, double zeta, double* GEOM,Vec3* gder4);

	//************************************
	// Method:    get_Grad_N4_flat
	// FullName:  femlib::get_Grad_N4_flat
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: get Grad of the interpolated shapefunctions of the TET4Flat routine 
	// Parameter: double xi
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM          std. TET4 Array
	// Parameter: Vec3 * gder4
	//************************************
	_EXTERN_ double get_Grad_N4_flat(double xi, double eta, double zeta, double* GEOM,Vec3* gder4);

	//************************************
	// Method:    get_Grad_N4_forInitFunction
	// FullName:  femlib::get_Grad_N4_forInitFunction
	// Access:    public 
	// Returns:   _EXTERN_ double
	// Description: get Grad of the interpolated shapefunctions of the N4 routine. Needs some special initalisation in HadaptSE
	// Parameter: double xi
	// Parameter: double eta
	// Parameter: double zeta
	// Parameter: double * GEOM
	// Parameter: Vec3 * gder4
	//************************************
	_EXTERN_ double get_Grad_N4_forInitFunction(double xi, double eta, double zeta, double* GEOM,Vec3* gder4);

	//************************************
	// Method:    get_GradN4_QP
	// FullName:  femlib::get_GradN4_QP
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description:  get Grad of the interpolated shapefunctions of the N4 routine at the QP. grad[i] for qp_i. 
	//				The same QPs as in the QTET4 Routine are used
	// Parameter: double * GEOM
	// Parameter: Vec3 *  & gradN4
	//************************************
	_EXTERN_ void get_GradN4_QP(double* GEOM, Vec3* &gradN4);

	//************************************
	// Method:    get_GradN4_QP_flat
	// FullName:  femlib::get_GradN4_QP_flat
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: et Grad of the interpolated shapefunctions of the TET4Flat routine at the QP. grad[i] for qp_i. 
	//				The same QPs as in the QTET4 Routine are used
	// Parameter: double * GEOM				std. TET4 GEOM Array
	// Parameter: Vec3 *  & gradN4
	//************************************
	_EXTERN_ void get_GradN4_QP_flat(double* GEOM, Vec3* &gradN4);



	//************************************
	// Method:    gradHat_centerOfGravity
	// FullName:  femlib::gradHat_centerOfGravity
	// Access:    public 
	// Returns:   _EXTERN_ Vec3
	// Description: get the gradient at the center of gravity (xi=eta=zeta=0.0) of an tet 10
	// Parameter: double * GEOM			std. TET10 GEOM Array
	// Parameter: unsigned i
	//************************************
	_EXTERN_ Vec3 gradHat_centerOfGravity(double* GEOM, unsigned i );

	//************************************
	// Method:    gradHat_centerOfGravity
	// FullName:  femlib::gradHat_centerOfGravity
	// Access:    public 
	// Returns:   _EXTERN_ Vec3
	// Description: get the gradient at the center of gravity (xi=eta=zeta=0.0) of an tet 4
	// Parameter: double * GEOM			std. TET4 GEOM Array
	// Parameter: unsigned i
	//************************************
	_EXTERN_ Vec3 gradHat_centerOfGravity_flat(double* GEOM, unsigned i );
} // end femlib

#endif // GradFunctions_h__