
#ifndef Mappings_h__
#define Mappings_h__
#include "defines.h"

namespace femlib
{
	///! Map from Node4 to Node10
	//!0<local_node4 && local_node4<5
	//************************************
	// Method:    Node4_2_Node10
	// FullName:  femlib::Node4_2_Node10
	// Access:    public 
	// Returns:   _EXTERN_ unsigned
	// Description: Convert from local node numbering of an tet4 to an local node numbering of an tet10
	// Parameter: unsigned local_node4
	//************************************
	_EXTERN_ unsigned Node4_2_Node10( unsigned local_node4);

	//! Map from Node10 to Node4, if auxnode return 0
	//************************************
	// Method:    Node10_2_Node4
	// FullName:  femlib::Node10_2_Node4
	// Access:    public 
	// Returns:   _EXTERN_ unsigned
	// Description: Convert from local node numbering of an tet10 to an local node numbering of an tet4
	// Parameter: unsigned local_node10
	//************************************
	_EXTERN_ unsigned Node10_2_Node4( unsigned local_node10) ;


	//Rotating
	//FK-rm: array adr+1
	//************************************
	// Method:    rotate_mapping
	// FullName:  femlib::rotate_mapping
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: Rotates the local node numbering, so that "face" has always the number 1,2,3,4,5,6
	// Parameter: unsigned face
	// Parameter: unsigned * rotate
	//************************************
	_EXTERN_ void rotate_mapping(unsigned face, unsigned* rotate);
	//************************************
	// Method:    rotate_edge_mapping
	// FullName:  femlib::rotate_edge_mapping
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: See roate_mapping, just with the edges
	// Parameter: unsigned face
	// Parameter: unsigned * rotate
	//************************************
	_EXTERN_ void rotate_edge_mapping( unsigned face, unsigned* rotate );
	//************************************
	// Method:    inv_rotate_mapping
	// FullName:  femlib::inv_rotate_mapping
	// Access:    public 
	// Returns:   _EXTERN_ void
	// Description: The inverse mapping to the rotate_mapping 
	// Parameter: unsigned was_face
	// Parameter: unsigned * rotate
	//************************************
	_EXTERN_ void inv_rotate_mapping( unsigned was_face, unsigned* rotate );
}





#endif // Mappings_h__

