#ifndef COMPLEX_H
#define COMPLEX_H

#include <cmath>
#include <iostream>

struct scomp {
  float re, im;
  scomp() {}
  scomp(float x, float y) : re(x), im(y) {}
  scomp(float x): re(x), im(0.0) {}
};

struct dcomp {
  double re, im;
  dcomp() {}
  dcomp(double x, double y) : re(x), im(y) {}
  dcomp(double x): re(x), im(0.0) {}
};

// Conjugate
inline scomp conj(scomp a) { a.im = -a.im; return a; }
inline dcomp conj(dcomp a) { a.im = -a.im; return a; }
inline scomp nconj(scomp a) { a.re = -a.re; return a; }
inline dcomp nconj(dcomp a) { a.re = -a.re; return a; }
inline double abs2(dcomp a) { return a.re*a.re + a.im*a.im; }
inline float abs2(scomp a) { return a.re*a.re + a.im*a.im; }
inline double abs(dcomp a) { return sqrt(abs2(a)); }
inline float abs(scomp a) { return sqrtf(abs2(a)); }
inline scomp operator-(scomp a) { a.re = -a.re; a.im = -a.im; return a; }
inline dcomp operator-(dcomp a) { a.re = -a.re; a.im = -a.im; return a; }

// Additionsoperatoren


inline scomp operator+=(scomp& a, scomp b)
{
  a.re += b.re;
  a.im += b.im;
  return a;
}

inline scomp operator+(scomp a, scomp b) { return (a += b); }

inline dcomp operator+=(dcomp& a, dcomp b)
{
  a.re += b.re;
  a.im += b.im;
  return a;
}

inline dcomp operator+=(dcomp& a, scomp b)
{
  a.re += b.re;
  a.im += b.im;
  return a;
}

inline dcomp operator+(dcomp a, dcomp b) { return (a += b); }

inline scomp operator+(scomp a, float b) { a.re += b; return a; }
inline scomp operator+=(scomp& a, float b) { a.re += b; return a; }
inline scomp operator+(float b, scomp a) { a.re += b; return a; }
inline dcomp operator+(dcomp a, double b) { a.re += b; return a; }
inline dcomp operator+=(dcomp& a, double b) { a.re += b; return a; }
inline dcomp operator+(double b, dcomp a) { a.re += b; return a; }

//Subtraktionsoperator
inline scomp operator-(scomp a, scomp b)
{
  a.re -= b.re;
  a.im -= b.im;
  return a;
}

inline scomp operator-=(scomp& a, scomp b)
{
  a.re -= b.re;
  a.im -= b.im;
  return a;
}

inline dcomp operator-(dcomp a, dcomp b)
{
  a.re -= b.re;
  a.im -= b.im;
  return a;
}

inline dcomp operator-=(dcomp& a, dcomp b)
{
  a.re -= b.re;
  a.im -= b.im;
  return a;
}

inline scomp operator-(scomp a, float b) { a.re -= b; return a; }
inline scomp operator-=(scomp& a, float b) { a.re -= b; return a; }
inline dcomp operator-(dcomp a, double b) { a.re -= b; return a; }
inline dcomp operator-=(dcomp& a, double b) { a.re -= b; return a; }


inline dcomp operator-(double b, dcomp a)
{
  dcomp r;
  r.re = b - a.re;
  r.im = -a.im;
  return r;
}

inline scomp operator-(float b, scomp a)
{
  scomp r;
  r.re = b - a.re;
  r.im = -a.im;
  return r;
}


// Multiplikationoperatoren
inline scomp operator*(scomp a, scomp b)
{
  scomp r;
  r.re = a.re*b.re - a.im*b.im;
  r.im = a.re*b.im + a.im*b.re;
  return r;
}


inline scomp operator*=(scomp& a, scomp b)
{
  float t = a.re*b.re - a.im*b.im;
  a.im *= b.re;
  a.im += a.re*b.im;
  a.re = t;
  return a;
}

inline dcomp operator*(dcomp a, dcomp b)
{
  dcomp r;
  r.re = a.re*b.re - a.im*b.im;
  r.im = a.re*b.im + a.im*b.re;
  return r;
}

inline dcomp operator*=(dcomp& a, dcomp b)
{
  double t = a.re*b.re - a.im*b.im;
  a.im *= b.re;
  a.im += a.re*b.im;
  a.re = t;
  return a;
}

inline scomp operator*(scomp a, float b) { a.re *= b; a.im *= b; return a; }
inline scomp operator*(float b, scomp a) { a.re *= b; a.im *= b; return a; }
inline dcomp operator*(dcomp a, double b) { a.re *= b; a.im *= b; return a; }
inline dcomp operator*(double b, dcomp a) { a.re *= b; a.im *= b; return a; }

// Divisionsoperator
inline scomp operator/(scomp a, scomp b)
{
  scomp r;
  r.re = (a.re*b.re + a.im*b.im)/abs2(b);
  r.im = (a.im*b.re - a.re*b.im)/abs2(b);
  return r;
}

inline dcomp operator/(dcomp a, dcomp b)
{
  dcomp r;
  r.re = (a.re*b.re + a.im*b.im)/abs2(b);
  r.im = (a.im*b.re - a.re*b.im)/abs2(b);
  return r;
}

inline dcomp operator/(double a, dcomp b)
{
  double d = a/abs2(b);
  b.re *=  d;
  b.im *= -d;
  return b;
}

inline scomp operator/(float a, scomp b)
{
  float d = a/abs2(b);
  b.re *=  d;
  b.im *= -d;
  return b;
}

inline dcomp operator/(dcomp a, double b)
{
  double d = 1.0/b;
  a.re *= d;
  a.im *= d;
  return a;
}

inline scomp operator/(scomp a, float b)
{
  float d = (float) 1.0/b;
  a.re *= d;
  a.im *= d;
  return a;
}


// Ausgabe
inline std::ostream& operator<<(std::ostream& os, dcomp z)
{ return os << '(' << z.re << ", " << z.im << ')'; }

inline std::ostream& operator<<(std::ostream& os, scomp z)
{ return os << '(' << z.re << ", " << z.im << ')'; }


// Convert

inline void Convert_z2c(dcomp z, scomp& c)
{
  c.re=(float) z.re;
  c.im=(float) z.im;
}

#endif
