/********************************************************************
	created:	2008/08/27
	created:	27:8:2008   17:23
	filename: 	d:\Development\FEMLib++\FEMLib++\FEMLib++\Header\femlib.h
	file path:	d:\Development\FEMLib++\FEMLib++\FEMLib++\Header
	file base:	femlib
	file ext:	h
	author:		Fk
	
	purpose:	Central FEMLIB Headerfile. Includes all other Headerfiles. 
*********************************************************************/

#ifndef fe_h__
#define fe_h__

#include "GeometryInformations.h"
#include "Mappings.h"
#include "BLF.h"
#include "defines.h"
#include "GradFunctions.h"
#include "CurlFunctions.h"
#include "Shapefunctions.h"

//TET10 Memory
//memory for local shape function
extern double* FUN10;
//memory for local derivatives
extern double* LDER10;
//memory for global derivatives
extern double* GDER10;
//memory for jacobian inverse
extern double* JACIN;
//memory for jacobian
extern double* JAC;
//memory for  global derivates as vec3
extern Vec3*   GRAD10;
//TET4 Memory
//memory for shape function for tet4
extern double* FUN4;
//memory for local derivatives for tet4
extern double* LDER4;
//memory for global derivatives for tet4
extern double* GDER4;
//memory for global derivatives for tet4 as vec3
extern Vec3*   GRAD4;


//Edge Element
//Memory for signs for the global orientation of the edge elements
extern double* SIGNS;
//the node-labels for the global orientation of the edge elements (to calculate the signs)
extern unsigned* NODE_LABELS;
//the edge function
extern Vec3* EDGEFUNC;
extern Vec3* EDGEFUNC_QP;

//the curls of the edge functions
extern Vec3* CURL_EDGE_FUNC;
extern Vec3* CURL_LOCAL_2;

//Quadratrue
//the quadrature point positions
extern double* ABSS;
//the quadrature weights
extern double* WGHT;
//for the jacobian deterimant for the integration
extern double* JAC_DETS;

//Geoemtry
//the std. geom to store the node coordinate of the tet10
extern double* GEOM;


namespace femlib
{
//////////////////////////////////////////////////////////////////////////
// Shape function
//////////////////////////////////////////////////////////////////////////
//!curved second order TET (TET10)
//************************************
// Method:    Tet10
// FullName:  femlib::Tet10
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Returns the nodal Shapefunctions FUN and the local derivatives LDER of a curved TET10 
//				at the point (xi,eta,zeta). The center of gravity is at xi=eta=zeta=0.0
// Parameter: double * FUN	(size:10)		(Output: Shapefunctions)
// Parameter: double * LDER	(size:3*10)		(Output: Local Derivaties of Shaepfunctions)
// Parameter: const double xi		
// Parameter: const double eta
// Parameter: const double zeta
//************************************
_EXTERN_ void Tet10(double* FUN, double* LDER,const double xi, const double eta, const double zeta);
//************************************


// Method:    Tet10
// FullName:  femlib::Tet10
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Returns the nodal Shapefunctions FUN for a curved TET10 
//				at the point (xi,eta,zeta). 
//				The center of gravity is at xi=eta=zeta=0.0
// Parameter: double * FUN	(size:10)		(Output: Shapefunctions)
// Parameter: const double xi		
// Parameter: const double eta
// Parameter: const double zeta
//************************************
_EXTERN_ void Tet10(double* FUN, const double xi, const double eta, const double zeta);

//!flat, first order TET	(TET4)
//************************************
// Method:    Tet4
// FullName:  femlib::Tet4
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Returns the nodal Shapefunctions for an flat TET4 
//				at the point (xi,eta,zeta). 
//				The center of gravity is at xi=eta=zeta=0.0
// Parameter: double * FUN		size:4		(Output: Shapefunctions)
// Parameter: double * LDER		size:3*4		(Output: Local Derivaties of Shaepfunctions)
// Parameter: const double xi
// Parameter: const double eta
// Parameter: const double zeta
//************************************
_EXTERN_ void Tet4Flat(double* FUN, double* LDER, const double xi, const double eta, const double zeta);
_EXTERN_ void Tet4Flat(double* FUN, const double xi, const double eta, const double zeta);

//!curved, second order triangle (TRIM6)
//************************************
// Method:    Trim6
// FullName:  femlib::Trim6
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Shapefunctions and there local derivates for the curved, secound order triangle (in 2D)
// Parameter: double * FUN	(size:6)		Shapefunctions
// Parameter: double * LDER	(size:2*6)		Local derivaties
// Parameter: double xi
// Parameter: double eta
//************************************
_EXTERN_ void Trim6(double* FUN, double* LDER, double xi, double eta);

//////////////////////////////////////////////////////////////////////////
//Quad Points
//////////////////////////////////////////////////////////////////////////
//FOUR TET quadrature points and weight 

//************************************
// Method:    QTET4
// FullName:  femlib::QTET4
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Quadrature Points (local coordinates) of the Tetrahedral reference element
// Parameter: double * WGHT		(size:4)	Weights. Same weight for all quad points
// Parameter: double * ABSS		(size:3*4)  local Coordinates for quadrature points
//************************************
_EXTERN_ void QTET4(double* WGHT, double* ABSS );

//************************************
// Method:    QTET4
// FullName:  femlib::QTET4
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Quadrature Points for the Tetrahedral element
// Parameter: double WGHT				Same weight for all quad points
// Parameter: double * ABSS		(size:3*4)  local Coordinates for quadrature points
//************************************
_EXTERN_ void QTET4(double WGHT, double* ABSS );

//SEVEN TRIA QP 
_EXTERN_ void QTRI7(double* WGHT, double* ABSS);
//Four TRIA QP
_EXTERN_ void QTRI4(double* WGHT, double* ABSS);
//////////////////////////////////////////////////////////////////////////
// Matrix Routines
//////////////////////////////////////////////////////////////////////////
//! Inverte 2x2 and 3x3 Matrices
//************************************
// Method:    MatInv
// FullName:  femlib::MatInv
// Access:    public 
// Returns:   _EXTERN_ int
// Description: Inversion routine for 2x2 or 3x3 matrices. Based on Cramers Rule
// Parameter: double * A	size:nxn		Inputmatrix. Stored in Fortran Format
// Parameter: unsigned n	n==2 || n==3
// Parameter: double * B	size:nxn		OutputMatrix. Stored in Fortran Format
// Parameter: double & det  Is the det of A. Usefull for jacobian determinant
//************************************
_EXTERN_ int MatInv(double* A, unsigned n, double* B, double &det);		//returns -1 if det == 0 and else 0

//************************************
// Method:    getDet
// FullName:  femlib::getDet
// Access:    public 
// Returns:   _EXTERN_ double
// Description: Calculates the det of the nxn Matrix A
// Parameter: double * A		Inputmatrix nxn
// Parameter: unsigned n		n==2 || n==3
//************************************
_EXTERN_ double getDet(double* A, unsigned n);			//Matrix deterinant

//! C=A*B        A in R^(mA,nA), B in R^(mA,nB) C in R^(mA,nB)
//************************************
// Method:    MatMul
// FullName:  femlib::MatMul
// Access:    public 
// Returns:   _EXTERN_ void
// Description: C=A*B			A in R^(mA,nA), B in R^(mA,nB) C in R^(mA,nB)
// Parameter: double * A		A in R^(mA,nA
// Parameter: unsigned mA
// Parameter: unsigned nA
// Parameter: double * B		B in R^(mA,nB)	
// Parameter: unsigned nB
// Parameter: double * C		C in R^(mA,nB)
//************************************
_EXTERN_ void MatMul(double* A, unsigned mA, unsigned nA, double* B, unsigned nB, double* C);

//! Matrix times vector  
//************************************
// Method:    MatVec
// FullName:  femlib::MatVec
// Access:    public 
// Returns:   _EXTERN_ void
// Description: Matrix-Vector Multiplication
// Parameter: double * A		A in R^(m,n)
// Parameter: unsigned m		
// Parameter: unsigned n
// Parameter: double * x		x in R^(n)
// Parameter: double * b		b in R^(m)
//************************************
_EXTERN_ void MatVec(double* A, unsigned m, unsigned n, double* x, double* b );

}//end femlib namespace



#endif // fe_h__