/*! \file vector3d.h
    \brief 3-Dim Vectors with Tools 
      
    Vectors of double or int with operations:
    + -  *  /  corss-product, dot-product etc, etc...  
*/

#ifndef VECTOR2D_
#define VECTOR2D_

//Includes
#include <iostream>
#include <complex>
#include <cassert>
using std::cout;
using std::endl;
namespace vector2d {


/** Templated Class for a simple vector with three components */
template <typename T>
class Vector2D
{
	private:
	T x[ 2 ]; 

	public:
		Vector2D( void ){}
		~Vector2D( void ){}

		Vector2D( const T &val ) 
			{ x[ 0 ] = x[ 1 ]= val;  }

		Vector2D( const Vector2D &v ) 
			{ x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];   }

		Vector2D( const T &x0,const T &x1 )
			{ x[ 0 ] = x0;  x[ 1 ] = x1;  }

		template <typename Y> Vector2D( const Y *a ) 
			{ x[ 0 ] = a[ 0 ];  x[ 1 ] = a[ 1 ];  }

		template <typename Y> Vector2D &operator = ( const Vector2D<Y> &v ) 
		{ x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];   return *this;  }
  
		operator T* () { return x;  }
		operator const T* () const { return x;  }
		T& operator [  ] ( int i ); 
		const T& operator [  ] ( int i ) const; 

		Vector2D  operator  - () const; 

  void output() const { 
	  std::cout << "[ " << x[ 0 ] << "," << x[ 1 ] << " ]" <<std::endl;  }
  double norm( void ) const { return sqrt( ( double )( x[ 0 ]*x[ 0 ]+x[ 1 ]*x[ 1 ] ) );  }
  int size( void ) { return 2;  }
  T get( int position ) const ; 
  void set( int position,  T val ); 
};   

template <typename T>
inline T Vector2D<T> :: get( int position ) const 
                   {assert( position>=0 && position<=1 ); 
                    return x[ position ]; }; 

template <typename T>
inline void Vector2D<T> :: set( int position, T val ) 
                   {assert( position>=0 && position<=1 ); 
                    x[ position ]=val; }; 

template <typename T> 
inline T& Vector2D<T>::operator [  ] ( int i )
{ assert( ( i>=0 ) && ( i<2 ) );  return x[ i ];  }

template <typename T> 
inline const T& Vector2D<T>::operator [  ] ( int i ) const
{ assert( ( i>=0 ) && ( i<2 ) );  return x[ i ];  }

template <typename T> 
inline Vector2D<T> Vector2D<T>::operator  - () const 
{ return Vector2D<T>( -x[ 0 ],-x[ 1 ] );  }

template<typename T> 
Vector2D<T> operator+( const Vector2D<T>  &a, const Vector2D<T>  &b )  
{return Vector2D<T>( a.get( 0 )+b.get( 0 ), a.get( 1 )+b.get( 1 ) ); }; 

template<typename T> 
Vector2D<T> operator-( const Vector2D<T>  &a, const Vector2D<T>  &b )   
{return Vector2D<T>( a.get( 0 )-b.get( 0 ), a.get( 1 )-b.get( 1 ) ); }; 


template<typename T> 
T operator*( const Vector2D<T>  &a, const Vector2D<T>  &b )   
{return T( a.get( 0 )*b.get( 0 ) + a.get( 1 )*b.get( 1 ) ); }; 

template<typename T> 
T operator*( const T  &a, const Vector2D<T>  &b )   
{return T( a*b.get( 0 ) + a*b.get( 1 ) ); }; 

template<typename T> 
Vector2D<T> operator*( const T  &b, const Vector2D<T>  &a ) 
{ return Vector2D<T> ( a.get( 0 )*b,a.get( 1 )*b );  } 

template<typename T> 
Vector2D<T> operator*(  const Vector2D<T>  &a,const T  &b ) 
{ return Vector2D<T> ( a.get( 0 )*b,a.get( 1 )*b );  } 
template<typename T> 
Vector2D<T> operator/( const Vector2D<T>  &a, const T  &b )  
{ return Vector2D<T>  ( a.get( 0 )/b,a.get( 1 )/b );  } 

typedef Vector2D<double> Vec2; 

//template<typename T> 
//typedef Vector2D<T> Vec3; 

}; 

/*
// #include <complex>

namespace vector3d {

  // Type promotion trait template.

  // Cf. Veldhuizen ``Techniques for scientific C++'' 
  template <class T1,class T2> struct promote_trait {}; 
  
#define DECLARE_PROMOTE( A,B,C ) \
template<> struct promote_trait<A,B> { typedef C T_promote;  };  

  DECLARE_PROMOTE( double,double,double ); 
  DECLARE_PROMOTE( double,double,double ); 
//  DECLARE_PROMOTE( double,complex<double>,complex<double> ); 
//   DECLARE_PROMOTE( complex<double>,double,complex<double> ); 
//   DECLARE_PROMOTE( double,complex<double>,complex<double> ); 
//   DECLARE_PROMOTE( complex<double>,double,complex<double> ); 
  DECLARE_PROMOTE( double,double,double ); 
//   DECLARE_PROMOTE( complex<double>,complex<double>,complex<double> ); 
  DECLARE_PROMOTE( double,double,double ); 
  DECLARE_PROMOTE( double,int,double ); 
  DECLARE_PROMOTE( int,double,double ); 
  DECLARE_PROMOTE( double,int,double ); 
  DECLARE_PROMOTE( int,double,double ); 

///  Templated Class for a simple vector with three components 
template <typename T>
class Vector2D
{
private:
  T x[ 3 ]; 
public:
  Vector2D( void ) {}
  Vector2D( const T &val )
  { x[ 0 ] = x[ 1 ] = x[ 2 ] = val;  }
  Vector2D( const Vector2D &v ) { x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  }
  template <typename Y> Vector2D( const Vector2D<Y> &v ) 
  { x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  }
  Vector2D( const T &x0,const T &x1,const T &x2 )
  { x[ 0 ] = x0;  x[ 1 ] = x1;  x[ 2 ] = x2;  }
  template <typename Y> Vector2D( const Y *a ) 
  { x[ 0 ] = a[ 0 ];  x[ 1 ] = a[ 1 ];  x[ 2 ] = a[ 2 ];  }
  template <typename Y> Vector2D &operator = ( const Vector2D<Y> &v ) 
  { x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  return *this;  }
  template <typename Y> bool operator == ( const Vector2D<Y> &b ) const; 
  
  operator T* () { return x;  }
  operator const T* () const { return x;  }
  T& operator [  ] ( int i ); 
  const T& operator [  ] ( int i ) const; 

  template <typename N> Vector2D operator hihi ( Vector2D<N> a, Vector2D<N> b ); 

  Vector2D  operator  - () const; 
  template <typename T> Vector2D<typename promote_trait<T,T>::T_promote>  
  operator  + ( const Vector2D<T> &b ) const; 
  template <typename Y> Vector2D<typename promote_trait<T,Y>::T_promote>  
  operator  - ( const Vector2D<Y> &b ) const; 
  template <typename N> Vector2D<typename promote_trait<T,N>::T_promote>
  operator * ( N f ) const; 
  template <typename N> Vector2D<typename promote_trait<T,N>::T_promote>
  operator / ( N f ) const; 
  template <typename N> Vector2D &operator *= ( N f ); 
  template <typename N> Vector2D &operator /= ( N f ); 
  template <typename Y> Vector2D &operator += ( const Vector2D<Y> &c ); 
  template <typename Y> Vector2D &operator -= ( const Vector2D<Y> &c ); 
  double norm( void ) const { return sqrt( ( double )( x[ 0 ]*x[ 0 ]+x[ 1 ]*x[ 1 ]+x[ 2 ]*x[ 2 ] ) );  }
  int size( void ) { return 3;  }
  T get( int position ) const ; 
  void set( int position,  T val ); 

//  friend ostream &operator << <> ( ostream &,const Vector2D &v ); 
  /// Cross product.
  template <typename Y> Vector2D<typename promote_trait<T,Y>::T_promote>
  operator % ( const Vector2D<Y> &b ) const; 
};   

template <typename T>
inline T Vector2D<T> :: get( int position ) const 
                   {assert( position>=0 && position<=2 ); 
                    return x[ position ]; }; 
template <typename T>
inline void Vector2D<T> :: set( int position, T val ) 
                   {assert( position>=0 && position<=2 ); 
                    x[ position ]=val; }; 

//template<typename T>
//inline ostream &operator << ( ostream &out,const Vector2D<T> &v )
//{ return out << '[ ' << v.x[ 0 ] << ',' << v.x[ 1 ] << ',' << v.x[ 2 ] << ' ]';  }

template<typename T> 
inline bool Vector2D<T>::operator == ( const Vector2D<T> &b ) const
{ return ( ( x[ 0 ] == b.x[ 0 ] ) && ( x[ 1 ] == b.x[ 1 ] ) && ( x[ 2 ] == b.x[ 2 ] ) );  }

template <typename T> 
inline T& Vector2D<T>::operator [  ] ( int i )
{ assert( ( i>=0 ) && ( i<3 ) );  return x[ i ];  }

template <typename T> 
inline const T& Vector2D<T>::operator [  ] ( int i ) const
{ assert( ( i>=0 ) && ( i<3 ) );  return x[ i ];  }

template <typename T> 
inline Vector2D<T> Vector2D<T>::operator  - () const 
{ return Vector2D( -x[ 0 ],-x[ 1 ],-x[ 2 ] );  }

template <typename T> 
inline Vector2D<typename promote_trait<T,T>::T_promote>
Vector2D<T>::operator  + ( const Vector2D<T> &b ) const
{ return Vector2D<typename promote_trait<T,T>::T_promote>
    ( x[ 0 ]+b[ 0 ],x[ 1 ]+b[ 1 ],x[ 2 ]+b[ 2 ] );  }

template <typename T> template <typename Y>
inline Vector2D<typename promote_trait<T,Y>::T_promote>
Vector2D<T>::operator  - ( const Vector2D<Y> &b ) const
{ return Vector2D<typename promote_trait<T,Y>::T_promote>
    ( x[ 0 ]-b[ 0 ],x[ 1 ]-b[ 1 ],x[ 2 ]-b[ 2 ] );  }

template <typename T> template <typename N>
inline  Vector2D<typename promote_trait<T,N>::T_promote>
Vector2D<T>::operator * ( N f ) const
{ return Vector2D<typename promote_trait<T,N>::T_promote>
    ( x[ 0 ]*f,x[ 1 ]*f,x[ 2 ]*f );  }

template <typename T> template <typename N>
inline  Vector2D<typename promote_trait<T,N>::T_promote>
Vector2D<T>::operator / ( N f ) const
{ return Vector2D<typename promote_trait<T,N>::T_promote>
    ( x[ 0 ]/f,x[ 1 ]/f,x[ 2 ]/f );  }


///  Prefix scalar multiplication.

This procedure is haunted by a problem:
\begin{verbatim}
  Vector2D<complex<double> > v( 1,1,1 ); 
  double s = 2.0; 
  cout << s*v; 
\end{verbatim}
lead to an error
\begin{verbatim}
coord.h:xxx: no matching function for call to 
`grid::Vector2D<double>::Vector2D ( complex<double>, complex<double>, complex<double> )'
\end{verbatim}
because of the attempt to construct a \textsf{Vector2D<double>} from a 3-tupel
of complex numbers. Remedy might be the trait mediated implicit type conversion
proposed by Veldhuizen. 

template <typename N,typename T>
inline Vector2D<typename promote_trait<T,N>::T_promote>
operator * ( N f,const Vector2D<T> &x )
{ return Vector2D<typename promote_trait<T,N>::T_promote>
    ( x[ 0 ]*f,x[ 1 ]*f,x[ 2 ]*f );  } 

/// Dot product.
template <typename T, typename Y>
inline typename promote_trait<T,Y>::T_promote 
dot( const Vector2D<T> &a,const Vector2D<Y> &b )
{ return ( a[ 0 ]*b[ 0 ]+a[ 1 ]*b[ 1 ]+a[ 2 ]*b[ 2 ] );  }

/// Cross product ( operator % )
template <typename T> template <typename Y> 
inline Vector2D<typename promote_trait<T,Y>::T_promote>
Vector2D<T>::operator % ( const Vector2D<Y> &b ) const
{ return Vector2D<typename promote_trait<T,Y>::T_promote>
    ( x[ 1 ]*b[ 2 ]-x[ 2 ]*b[ 1 ],x[ 2 ]*b[ 0 ]-x[ 0 ]*b[ 2 ],x[ 0 ]*b[ 1 ]-x[ 1 ]*b[ 0 ] );  }

template <typename T> template <typename N>
inline Vector2D<T> &Vector2D<T>::operator *= ( N f )
{ x[ 0 ] *= f;  x[ 1 ] *= f;  x[ 2 ] *= f;  return *this;  }

template <typename T> template <typename N>
inline Vector2D<T> &Vector2D<T>::operator /= ( N f )
{ x[ 0 ] *= f;  x[ 1 ] *= f;  x[ 2 ] *= f;  return *this;  }

template <typename T> template <typename Y>
inline Vector2D<T> &Vector2D<T>::operator += ( const Vector2D<Y> &c )
{ x[ 0 ] += c[ 0 ];  x[ 1 ] += c[ 1 ];  x[ 2 ] += c[ 2 ];  return *this;  }

template <typename T> template <typename Y>
inline Vector2D<T> &Vector2D<T>::operator -= ( const Vector2D<Y> &c )
{ x[ 0 ] -= c[ 0 ];  x[ 1 ] -= c[ 1 ];  x[ 2 ] -= c[ 2 ];  return *this;  }

typedef Vector2D<double> Vec3; 

}; 

*/

#endif
