/*! \file vector3d.h
    \brief 3-Dim Vectors with Tools 
      
    Vectors of double or int with operations:
    + -  *  /  corss-product, dot-product etc, etc...  
*/

#ifndef VECTOR3D_
#define VECTOR3D_

//Includes
//#include <complex>
#include <iostream>
#include "complex_hadapt.h"
#include <cassert>
using std::cout;
using std::endl;
namespace vector3d {


/** Templated Class for a simple vector with three components */
template <typename T>
class Vector3D
{
	private:
	T x[ 3 ]; 

	public:
		Vector3D( void ){}
		~Vector3D( void ){}

		Vector3D( const T &val ) 
			{ x[ 0 ] = x[ 1 ] = x[ 2 ] = val;  }

		Vector3D( const Vector3D &v ) 
			{ x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  }

		Vector3D( const T &x0,const T &x1,const T &x2 )
			{ x[ 0 ] = x0;  x[ 1 ] = x1;  x[ 2 ] = x2;  }

		template <typename Y> Vector3D( const Y *a ) 
			{ x[ 0 ] = a[ 0 ];  x[ 1 ] = a[ 1 ];  x[ 2 ] = a[ 2 ];  }

		template <typename Y> Vector3D &operator = ( const Vector3D<Y> &v ) 
		{ x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  return *this;  }
  
		operator T* () { return x;  }
		operator const T* () const { return x;  }
		T& operator [  ] ( int i ); 
		const T& operator [  ] ( int i ) const; 

		Vector3D  operator  - () const; 
	//! Prints the real stored values, instead of the cleaned ones in outputc	
  void output() const { 

	  std::cout << "[ " << x[ 0 ] << "," << x[ 1 ] << "," << x[ 2 ] << " ]" <<std::endl;  }
  //!Print instead of small values 0.0 ... just looks nicer.
   //************************************
   // Method		: outputc
   // Description	: Replace all values witih fabs(value)<1.0e-16 with 0.0. Assumend that this is machine precision.
    //************************************
   void outputc() const { 
		double x1=x[0],x2=x[1],x3=x[2];
		if(fabs(x1)<1.0e-15) x1=0.0;
		if(fabs(x2)<1.0e-15) x2=0.0;
		if(fabs(x3)<1.0e-15) x3=0.0;
	  std::cout << "[ " << x1 << "," << x2 << "," << x3 << " ]" <<std::endl;  }

  double norm( void ) const { return sqrt( ( double )( x[ 0 ]*x[ 0 ]+x[ 1 ]*x[ 1 ]+x[ 2 ]*x[ 2 ] ) );  }
  int size( void ) { return 3;  }
  T get( int position ) const ; 
  void set( int position,  T val ); 
};   

template <typename T>
inline T Vector3D<T> :: get( int position ) const 
                   {assert( position>=0 && position<=2 ); 
                    return x[ position ]; }; 

template <typename T>
inline void Vector3D<T> :: set( int position, T val ) 
                   {assert( position>=0 && position<=2 ); 
                    x[ position ]=val; }; 

template <typename T> 
inline T& Vector3D<T>::operator [  ] ( int i )
{ assert( ( i>=0 ) && ( i<3 ) );  return x[ i ];  }

template <typename T> 
inline const T& Vector3D<T>::operator [  ] ( int i ) const
{ assert( ( i>=0 ) && ( i<3 ) );  return x[ i ];  }

template <typename T> 
inline Vector3D<T> Vector3D<T>::operator  - () const 
{ return Vector3D<T>( -x[ 0 ],-x[ 1 ],-x[ 2 ] );  }

template<typename T> 
Vector3D<T> operator+( const Vector3D<T>  &a, const Vector3D<T>  &b )  
{return Vector3D<T>( a.get( 0 )+b.get( 0 ), a.get( 1 )+b.get( 1 ), a.get( 2 )+b.get( 2 ) ); }; 

template<typename T> 
Vector3D<T> operator-( const Vector3D<T>  &a, const Vector3D<T>  &b )   
{return Vector3D<T>( a.get( 0 )-b.get( 0 ), a.get( 1 )-b.get( 1 ), a.get( 2 )-b.get( 2 ) ); }; 


template<typename T> 
T operator*( const Vector3D<T>  &a, const Vector3D<T>  &b )   
{return T( a.get( 0 )*b.get( 0 ) + a.get( 1 )*b.get( 1 ) + a.get( 2 )*b.get( 2 ) ); }; 
//Cross Product
template<typename T> 
Vector3D<T> operator%( const Vector3D<T>  &a, const Vector3D<T>  &b )  
{ return Vector3D<T>
    ( a.get( 1 )*b.get( 2 )-a.get( 2 )*b.get( 1 ),
	 a.get( 2 )*b.get( 0 )-a.get( 0 )*b.get( 2 ),
	 a.get( 0 )*b.get( 1 )-a.get( 1 )*b.get( 0 ) );  }

template<typename T> 
Vector3D<T> operator*( const T  &b, const Vector3D<T>  &a ) 
{ return Vector3D<T> ( a.get( 0 )*b,a.get( 1 )*b,a.get( 2 )*b );  } 

template<typename T> 
Vector3D<T> operator/( const Vector3D<T>  &a, const T  &b )  
{ return Vector3D<T>  ( a.get( 0 )/b,a.get( 1 )/b,a.get( 2 )/b );  } 

template<typename T> 
bool operator==(const Vector3D<T> &a,const Vector3D<T> &b)
{
	return a.get(0)==b.get(0)&&a.get(1)==b.get(1)&&a.get(2)==b.get(2);
}

typedef Vector3D<double> Vec3; 
typedef Vector3D<dcomp> cVec3; 


//template<typename T> 
//typedef Vector3D<T> Vec3; 

}; 

/*
// #include <complex>

namespace vector3d {

  // Type promotion trait template.

  // Cf. Veldhuizen ``Techniques for scientific C++'' 
  template <class T1,class T2> struct promote_trait {}; 
  
#define DECLARE_PROMOTE( A,B,C ) \
template<> struct promote_trait<A,B> { typedef C T_promote;  };  

  DECLARE_PROMOTE( double,double,double ); 
  DECLARE_PROMOTE( double,double,double ); 
//  DECLARE_PROMOTE( double,complex<double>,complex<double> ); 
//   DECLARE_PROMOTE( complex<double>,double,complex<double> ); 
//   DECLARE_PROMOTE( double,complex<double>,complex<double> ); 
//   DECLARE_PROMOTE( complex<double>,double,complex<double> ); 
  DECLARE_PROMOTE( double,double,double ); 
//   DECLARE_PROMOTE( complex<double>,complex<double>,complex<double> ); 
  DECLARE_PROMOTE( double,double,double ); 
  DECLARE_PROMOTE( double,int,double ); 
  DECLARE_PROMOTE( int,double,double ); 
  DECLARE_PROMOTE( double,int,double ); 
  DECLARE_PROMOTE( int,double,double ); 

///  Templated Class for a simple vector with three components 
template <typename T>
class Vector3D
{
private:
  T x[ 3 ]; 
public:
  Vector3D( void ) {}
  Vector3D( const T &val )
  { x[ 0 ] = x[ 1 ] = x[ 2 ] = val;  }
  Vector3D( const Vector3D &v ) { x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  }
  template <typename Y> Vector3D( const Vector3D<Y> &v ) 
  { x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  }
  Vector3D( const T &x0,const T &x1,const T &x2 )
  { x[ 0 ] = x0;  x[ 1 ] = x1;  x[ 2 ] = x2;  }
  template <typename Y> Vector3D( const Y *a ) 
  { x[ 0 ] = a[ 0 ];  x[ 1 ] = a[ 1 ];  x[ 2 ] = a[ 2 ];  }
  template <typename Y> Vector3D &operator = ( const Vector3D<Y> &v ) 
  { x[ 0 ] = v[ 0 ];  x[ 1 ] = v[ 1 ];  x[ 2 ] = v[ 2 ];  return *this;  }
  template <typename Y> bool operator == ( const Vector3D<Y> &b ) const; 
  
  operator T* () { return x;  }
  operator const T* () const { return x;  }
  T& operator [  ] ( int i ); 
  const T& operator [  ] ( int i ) const; 

  template <typename N> Vector3D operator hihi ( Vector3D<N> a, Vector3D<N> b ); 

  Vector3D  operator  - () const; 
  template <typename T> Vector3D<typename promote_trait<T,T>::T_promote>  
  operator  + ( const Vector3D<T> &b ) const; 
  template <typename Y> Vector3D<typename promote_trait<T,Y>::T_promote>  
  operator  - ( const Vector3D<Y> &b ) const; 
  template <typename N> Vector3D<typename promote_trait<T,N>::T_promote>
  operator * ( N f ) const; 
  template <typename N> Vector3D<typename promote_trait<T,N>::T_promote>
  operator / ( N f ) const; 
  template <typename N> Vector3D &operator *= ( N f ); 
  template <typename N> Vector3D &operator /= ( N f ); 
  template <typename Y> Vector3D &operator += ( const Vector3D<Y> &c ); 
  template <typename Y> Vector3D &operator -= ( const Vector3D<Y> &c ); 
  double norm( void ) const { return sqrt( ( double )( x[ 0 ]*x[ 0 ]+x[ 1 ]*x[ 1 ]+x[ 2 ]*x[ 2 ] ) );  }
  int size( void ) { return 3;  }
  T get( int position ) const ; 
  void set( int position,  T val ); 

//  friend ostream &operator << <> ( ostream &,const Vector3D &v ); 
  /// Cross product.
  template <typename Y> Vector3D<typename promote_trait<T,Y>::T_promote>
  operator % ( const Vector3D<Y> &b ) const; 
};   

template <typename T>
inline T Vector3D<T> :: get( int position ) const 
                   {assert( position>=0 && position<=2 ); 
                    return x[ position ]; }; 
template <typename T>
inline void Vector3D<T> :: set( int position, T val ) 
                   {assert( position>=0 && position<=2 ); 
                    x[ position ]=val; }; 

//template<typename T>
//inline ostream &operator << ( ostream &out,const Vector3D<T> &v )
//{ return out << '[ ' << v.x[ 0 ] << ',' << v.x[ 1 ] << ',' << v.x[ 2 ] << ' ]';  }

template<typename T> 
inline bool Vector3D<T>::operator == ( const Vector3D<T> &b ) const
{ return ( ( x[ 0 ] == b.x[ 0 ] ) && ( x[ 1 ] == b.x[ 1 ] ) && ( x[ 2 ] == b.x[ 2 ] ) );  }

template <typename T> 
inline T& Vector3D<T>::operator [  ] ( int i )
{ assert( ( i>=0 ) && ( i<3 ) );  return x[ i ];  }

template <typename T> 
inline const T& Vector3D<T>::operator [  ] ( int i ) const
{ assert( ( i>=0 ) && ( i<3 ) );  return x[ i ];  }

template <typename T> 
inline Vector3D<T> Vector3D<T>::operator  - () const 
{ return Vector3D( -x[ 0 ],-x[ 1 ],-x[ 2 ] );  }

template <typename T> 
inline Vector3D<typename promote_trait<T,T>::T_promote>
Vector3D<T>::operator  + ( const Vector3D<T> &b ) const
{ return Vector3D<typename promote_trait<T,T>::T_promote>
    ( x[ 0 ]+b[ 0 ],x[ 1 ]+b[ 1 ],x[ 2 ]+b[ 2 ] );  }

template <typename T> template <typename Y>
inline Vector3D<typename promote_trait<T,Y>::T_promote>
Vector3D<T>::operator  - ( const Vector3D<Y> &b ) const
{ return Vector3D<typename promote_trait<T,Y>::T_promote>
    ( x[ 0 ]-b[ 0 ],x[ 1 ]-b[ 1 ],x[ 2 ]-b[ 2 ] );  }

template <typename T> template <typename N>
inline  Vector3D<typename promote_trait<T,N>::T_promote>
Vector3D<T>::operator * ( N f ) const
{ return Vector3D<typename promote_trait<T,N>::T_promote>
    ( x[ 0 ]*f,x[ 1 ]*f,x[ 2 ]*f );  }

template <typename T> template <typename N>
inline  Vector3D<typename promote_trait<T,N>::T_promote>
Vector3D<T>::operator / ( N f ) const
{ return Vector3D<typename promote_trait<T,N>::T_promote>
    ( x[ 0 ]/f,x[ 1 ]/f,x[ 2 ]/f );  }


///  Prefix scalar multiplication.

This procedure is haunted by a problem:
\begin{verbatim}
  Vector3D<complex<double> > v( 1,1,1 ); 
  double s = 2.0; 
  cout << s*v; 
\end{verbatim}
lead to an error
\begin{verbatim}
coord.h:xxx: no matching function for call to 
`grid::Vector3D<double>::Vector3D ( complex<double>, complex<double>, complex<double> )'
\end{verbatim}
because of the attempt to construct a \textsf{Vector3D<double>} from a 3-tupel
of complex numbers. Remedy might be the trait mediated implicit type conversion
proposed by Veldhuizen. 

template <typename N,typename T>
inline Vector3D<typename promote_trait<T,N>::T_promote>
operator * ( N f,const Vector3D<T> &x )
{ return Vector3D<typename promote_trait<T,N>::T_promote>
    ( x[ 0 ]*f,x[ 1 ]*f,x[ 2 ]*f );  } 

/// Dot product.
template <typename T, typename Y>
inline typename promote_trait<T,Y>::T_promote 
dot( const Vector3D<T> &a,const Vector3D<Y> &b )
{ return ( a[ 0 ]*b[ 0 ]+a[ 1 ]*b[ 1 ]+a[ 2 ]*b[ 2 ] );  }

/// Cross product ( operator % )
template <typename T> template <typename Y> 
inline Vector3D<typename promote_trait<T,Y>::T_promote>
Vector3D<T>::operator % ( const Vector3D<Y> &b ) const
{ return Vector3D<typename promote_trait<T,Y>::T_promote>
    ( x[ 1 ]*b[ 2 ]-x[ 2 ]*b[ 1 ],x[ 2 ]*b[ 0 ]-x[ 0 ]*b[ 2 ],x[ 0 ]*b[ 1 ]-x[ 1 ]*b[ 0 ] );  }

template <typename T> template <typename N>
inline Vector3D<T> &Vector3D<T>::operator *= ( N f )
{ x[ 0 ] *= f;  x[ 1 ] *= f;  x[ 2 ] *= f;  return *this;  }

template <typename T> template <typename N>
inline Vector3D<T> &Vector3D<T>::operator /= ( N f )
{ x[ 0 ] *= f;  x[ 1 ] *= f;  x[ 2 ] *= f;  return *this;  }

template <typename T> template <typename Y>
inline Vector3D<T> &Vector3D<T>::operator += ( const Vector3D<Y> &c )
{ x[ 0 ] += c[ 0 ];  x[ 1 ] += c[ 1 ];  x[ 2 ] += c[ 2 ];  return *this;  }

template <typename T> template <typename Y>
inline Vector3D<T> &Vector3D<T>::operator -= ( const Vector3D<Y> &c )
{ x[ 0 ] -= c[ 0 ];  x[ 1 ] -= c[ 1 ];  x[ 2 ] -= c[ 2 ];  return *this;  }

typedef Vector3D<double> Vec3; 

}; 

*/

#endif
